/*
 * ast.c
 *
 * Abstract syntax trees
 * Alloc/free routines
 */

char rcsid_ast[] = "$Id: ast.c,v 1.6 1992/08/14 00:01:17 pavel Exp $";

#include "my-string.h"

#include "ast.h"
#include "list.h"
#include "log.h"
#include "program.h"
#include "structures.h"
#include "sym_table.h"
#include "storage.h"
#include "utils.h"

static int	pool_size, next_pool_slot;
static void   **pool;

void
begin_code_allocation()
{
    pool_size = 10;
    next_pool_slot = 0;
    pool = (void **) mymalloc(pool_size * sizeof(void *), M_AST_POOL);
}

void
end_code_allocation(int aborted)
{
    if (aborted) {
	int	i;

	for (i = 0; i < next_pool_slot; i++) {
	    if (pool[i] != 0)
		myfree(pool[i], M_AST);
	}
    }

    myfree((void *) pool, M_AST_POOL);
}

void *
allocate(int size)
{
    if (next_pool_slot >= pool_size) { /* enlarge the pool */
	void  **new_pool;
	int	i;

	pool_size *= 2;
	new_pool = (void **) mymalloc(pool_size * sizeof(void *), M_AST_POOL);
	for (i = 0; i < next_pool_slot; i++) {
	    new_pool[i] = pool[i];
	}
	myfree((void *) pool, M_AST_POOL);
	pool = new_pool;
    }

    return pool[next_pool_slot++] = mymalloc(size, M_AST);
}

void
deallocate(void *ptr)
{
    int	i;

    for (i = 0; i < next_pool_slot; i++) {
	if (ptr == pool[i]) {
	    myfree(ptr, M_AST);
	    pool[i] = 0;
	    return;
	}
    }

    errlog("DEALLOCATE: Unknown pointer deallocated: 0x%x", (unsigned) ptr);
}

char *
ak_alloc_string(char *buffer)
{
    char	*string;

    string = (char *) mymalloc(sizeof(char) * (strlen(buffer) + 1), M_STRING);
    strcpy(string, buffer);
    return string;
}

/* added from 1.5 */
void 
ak_dealloc_string(char *string)
{
  myfree((void *)string, M_STRING);
}

Stmt *
alloc_stmt(enum Stmt_Kind kind)
{
    Stmt       *result = (Stmt *) allocate(sizeof(Stmt));

    result->kind = kind;
    result->next = (Stmt *) 0;
    return result;
}

Cond_Arm *
alloc_cond_arm(Expr *condition, Stmt *stmt)
{
    Cond_Arm   *result = (Cond_Arm *) allocate(sizeof(Cond_Arm));

    result->condition = condition;
    result->stmt = stmt;
    result->next = (Cond_Arm *) 0;
    return result;
}

Expr *
alloc_expr(enum Expr_Kind kind)
{
    Expr       *result = (Expr *) allocate(sizeof(Expr));
    
    result->kind = kind;
    return result;
}

Expr *
alloc_var(var_type type)
{
    Expr       *result = alloc_expr(EXPR_VAR);

    result->e.var.type = type;
    return result;
}

Expr *
alloc_binary(enum Expr_Kind kind, Expr *lhs, Expr *rhs)
{
    Expr       *result = alloc_expr(kind);
    
    result->e.bin.lhs = lhs;
    result->e.bin.rhs = rhs;
    return result;
}

Expr *
alloc_verb(Expr *obj, Expr *verb, Arg_List *args)
{
    Expr       *result = alloc_expr(EXPR_VERB);

    result->e.verb.obj = obj;
    result->e.verb.verb = verb;
    result->e.verb.args = args;
    return result;
}

Arg_List *
alloc_arg_list(enum Arg_Kind kind, Expr *expr)
{
    Arg_List   *result = (Arg_List *) allocate(sizeof(Arg_List));

    result->kind = kind;
    result->expr = expr;
    result->next = (Arg_List *) 0;
    return result;
}

Expr *
alloc_asgn(enum Asgn_Kind kind, Expr *value)
{
    Expr       *result = alloc_expr(EXPR_ASGN);

    result->e.asgn.kind = kind;
    result->e.asgn.value = value;
    return result;
}

static void free_expr(Expr *);

static void
free_arg_list(Arg_List *args)
{
    Arg_List   *arg, *next_arg;

    for (arg = args; arg; arg = next_arg) {
	next_arg = arg->next;
	free_expr(arg->expr);
	myfree((void *) arg, M_AST);
    }
}

static void
free_ast_var(Var v)
{
    int		i;
    
    switch ((int) v.type) {
      case TYPE_STR:
        if (v.v.str)
            myfree((void *) v.v.str, M_STRING);
	break;
      case TYPE_LIST:
	if(v.v.list) {
	    for(i = 1; i <= v.v.list[0].v.num; i++)
		free_ast_var( v.v.list[i]);
	    myfree((void *) v.v.list, M_LIST);
	}
	break;
      case TYPE_NUM:
      case TYPE_OBJ:
      case TYPE_ERR:
      case TYPE_ANY:
	break;
    }
}  

static void
free_expr(Expr *expr)
{
    switch (expr->kind) {

      case EXPR_VAR:
	free_ast_var(expr->e.var);
	break;

      case EXPR_ID:
	if (expr->e.id.name != NULL)
	    myfree((void *) expr->e.id.name, M_STRING);
	break;

      case EXPR_PROP:	case EXPR_INDEX:	case EXPR_PLUS:
      case EXPR_MINUS:  case EXPR_TIMES:	case EXPR_DIVIDE:
      case EXPR_MOD:	case EXPR_AND:		case EXPR_OR:
      case EXPR_EQ:     case EXPR_NE:		case EXPR_LT:
      case EXPR_LE:	case EXPR_GT:		case EXPR_GE:
      case EXPR_IN:
	free_expr(expr->e.bin.lhs);
	free_expr(expr->e.bin.rhs);
	break;

      case EXPR_COND:
	free_expr(expr->e.cond.condition);
	free_expr(expr->e.cond.consequent);
	free_expr(expr->e.cond.alternate);
	break;

      case EXPR_VERB:
	free_expr(expr->e.verb.obj);
	free_expr(expr->e.verb.verb);
	free_arg_list(expr->e.verb.args);
	break;

      case EXPR_RANGE:
	free_expr(expr->e.range.base);
	free_expr(expr->e.range.from);
	free_expr(expr->e.range.to);
	break;

      case EXPR_ASGN:
	if (expr->e.asgn.kind == ASGN_VAR) {
	    if (expr->e.asgn.a.id.name != NULL)
		myfree((void *) expr->e.asgn.a.id.name, M_STRING);
	} else {
	    free_expr(expr->e.asgn.a.bin.lhs);
	    free_expr(expr->e.asgn.a.bin.rhs);
	}
	free_expr(expr->e.asgn.value);
	break;

      case EXPR_CALL:
	free_arg_list(expr->e.call.args);
	break;

      case EXPR_NEGATE:
      case EXPR_NOT:
	free_expr(expr->e.expr);
	break;

      case EXPR_LIST:
	free_arg_list(expr->e.list);
	break;

      default:
	errlog("FREE_EXPR: Unknown Expr_Kind: %d\n", expr->kind);
	break;
    }

    myfree((void *) expr, M_AST);
}

void
free_stmt(Stmt *stmt)
{
    Stmt       *next_stmt;
    Cond_Arm   *arm, *next_arm;

    for (; stmt; stmt = next_stmt) {
	next_stmt = stmt->next;

	switch (stmt->kind) {

	  case STMT_COND:
	    for (arm = stmt->s.cond.arms; arm; arm = next_arm) {
		next_arm = arm->next;
		free_expr(arm->condition);
		free_stmt(arm->stmt);
		myfree((void *) arm, M_AST);
	    }
	    if (stmt->s.cond.otherwise)
		free_stmt(stmt->s.cond.otherwise);
	    break;

	  case STMT_LIST:
	    if (stmt->s.list.id.name != NULL)
		myfree((void *) stmt->s.list.id.name, M_STRING);
	    free_expr(stmt->s.list.expr);
	    free_stmt(stmt->s.list.body);
	    break;

	  case STMT_RANGE:
	    if (stmt->s.range.id.name != NULL)
		myfree((void *) stmt->s.range.id.name, M_STRING);
	    free_expr(stmt->s.range.from);
	    free_expr(stmt->s.range.to);
	    free_stmt(stmt->s.range.body);
	    break;

	  case STMT_WHILE:
	    free_expr(stmt->s.loop.condition);
	    free_stmt(stmt->s.loop.body);
	    break;

	  case STMT_FORK:
	    if (stmt->s.fork.id.name != NULL)
		myfree((void *) stmt->s.fork.id.name, M_STRING);
	    free_expr(stmt->s.fork.time);
	    free_stmt(stmt->s.fork.body);
	    break;

	  case STMT_EXPR:
	  case STMT_RETURN:
	    if (stmt->s.expr)
		free_expr(stmt->s.expr);
	    break;

	  default:
	    errlog("FREE_STMT: unknown Stmt_Kind: %d\n", stmt->kind);
	    break;
	}

	myfree((void *) stmt, M_AST);
    }
}

/* $Log: ast.c,v $
 * Revision 1.6  1992/08/14  00:01:17  pavel
 * Converted to a typedef of `var_type' = `enum var_type'.
 *
 * Revision 1.5  1992/08/10  16:53:07  pjames
 * Updated #includes.
 *
 * Revision 1.4  1992/07/30  21:20:25  pjames
 * Checks for NULL before freeing M_STRINGS.
 *
 * Revision 1.3  1992/07/27  17:57:06  pjames
 * Changed myfree(*, M_AST) to (*, M_STRING) when * was a (char *)
 * because that is how they are allocated.
 *
 * Revision 1.2  1992/07/20  23:48:48  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS identification string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 * */
