char rcsid_bf_verb[] = "$Id: bf_verb.c,v 1.3 1992/08/10 17:21:32 pjames Exp $";

#include "list.h"
#include "functions.h"
#include "parser.h"
#include "execute.h"
#include "unparse.h"
#include "log.h"
#include "my-string.h"
#include "utils.h"
#include "match.h"
#include "parse_command.h"
#include "storage.h"

/* internals */

static void
lister(void *data, const char *line)
{
    Var	*r = (Var *) data;
    Var	v;
    
    v.type = TYPE_STR;
    v.v.str = strdup(line);
    *r = listappend(*r, v);
}

static int
is_verb_info(Var *list)
{
    return (list[0].v.num == 3
	    && list[1].type == TYPE_OBJ
	    && list[2].type == TYPE_STR
	    && list[3].type == TYPE_STR);
}

static int
is_args(Var *list)
{
    return (list[0].v.num == 3
	    && list[1].type == TYPE_STR
	    && list[2].type == TYPE_STR
	    && list[3].type == TYPE_STR);
}

static int
is_code(Var *list)
{
    int		i;
    
    for (i = 1; i <= list[0].v.num; i++)
	if (list[i].type != TYPE_STR)
	    return 0;
    return 1;
}

static int
parse_verb_perms(const char *perm_str)
{
    int	new_perms = 0;
    
    while (*perm_str) {
	switch (*(perm_str++)) {
	  case 'r':
	  case 'R':
	    new_perms |= (1 << VP_READ);
	    break;
	  case 'w':
	  case 'W':
	    new_perms |= (1 << VP_WRITE);
	    break;
	  case 'x':
	  case 'X':
	    new_perms |= (1 << VP_EXEC);
	    break;
	  case 'd':
	  case 'D':
	    new_perms |= (1 << VP_DEBUG);
	    break;
	  default:
	    return -1;
	}
    }
    
    return new_perms;
}

struct parser_state {
    Var		code;		/* a list of strings */
    int		cur_string;	/* which string? */
    int		cur_char;	/* which character in that string? */
    Var		errors;		/* a list of strings */
};

static void
my_error(void *data, const char *msg)
{
    struct parser_state	*state = (struct parser_state *) data;
    Var			v;
    
    v.type = TYPE_STR;
    v.v.str = strdup(msg);
    state->errors = listappend(state->errors, v);
}

static int
my_getc(void *data)
{
    struct parser_state	*state = (struct parser_state *) data;
    Var			code;
    char		c;
    
    code = state->code;
    if (state->cur_string > code.v.list[0].v.num)
	return EOF;
    else if (!(c = code.v.list[state->cur_string].v.str[state->cur_char])) {
	state->cur_string++;
	state->cur_char = 0;
	return '\n';
    } else {
	state->cur_char++;
	return c;
    }
}

static void
my_ungetc(void *data, char c)
{
    struct parser_state	*state = (struct parser_state *) data;
    Var			code;
    
    code = state->code;
    if (state->cur_string == 1  &&  state->cur_char == 0)
	errlog("EVAL's UNGETC: At start of code\n");
    else {
	char	cc;
	
	if (state->cur_char > 0)
	    state->cur_char--;
	else {
	    state->cur_string--;
	    state->cur_char = strlen(code.v.list[state->cur_string].v.str);
	}
	cc = code.v.list[state->cur_string].v.str[state->cur_char];
	if (!(c == cc || (cc == '\0' && c == '\n')))
	    errlog("EVAL's UNGETC: Wrong char; c = %d and cc = %d\n", c, cc);
    }
}

static Parser_Client parser_client = { my_error, my_ungetc, my_getc };

/* externals */

static package
bf_verbs(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object) */
    Var r;
    Objid oid = arglist.v.list[1].v.obj;
    
    free_var(arglist);
    if (!retrieve(oid))
	return make_error_pack(E_INVARG);
    else if (!Obj_perm(progr, oid, FLAG_READ))
	return make_error_pack(E_PERM);
    else {
	Verbdef    *v;
	Var	    verb;
	int	    i, count = 0;
	
	for (v = retrieve(oid)->verbdefs; v; v = v->next)
	    count++;
	r = new_list(count);
	verb.type = TYPE_STR;
	for (v = retrieve(oid)->verbdefs, i = 1; v; v = v->next, i++) {
	    verb.v.str = strdup(v->name);
	    r.v.list[i] = verb;
	}
    }
    return make_var_pack(r);
}

static package
bf_verb_info(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, verb-name) */
    Var r;
    
    if (!retrieve(arglist.v.list[1].v.obj)) {
	free_var(arglist);
	return make_error_pack(E_INVARG);
    } else {
	int		vid, i = 0;
	Verbdef    *v = match_verbdef(arglist.v.list[2].v.str, 
				      arglist.v.list[1].v.obj, &vid);
	char       *s;
	
	if (!v) {
	    free_var(arglist);
	    return make_error_pack(E_VERBNF);
	} else if (!Verb_perm(progr, v, VP_READ)) { 
	    free_var(arglist);
	    return make_error_pack(E_PERM);
	} else {
	    r = new_list(3);
	    r.v.list[1].type = TYPE_OBJ;
	    r.v.list[1].v.obj = v->owner;
	    r.v.list[2].type = TYPE_STR;
	    r.v.list[2].v.str = s = strdup("xxxx");
	    if (v->perms & (1 << VP_READ)) s[i++] = 'r';
	    if (v->perms & (1 << VP_WRITE)) s[i++] = 'w';
	    if (v->perms & (1 << VP_EXEC)) s[i++] = 'x';
	    if (v->perms & (1 << VP_DEBUG)) s[i++] = 'd';
	    s[i] = '\0';
	    r.v.list[3].type = TYPE_STR;
	    r.v.list[3].v.str = strdup(v->name);
	}
    }
    
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_set_verb_info(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, verb-name, {owner, perms, name}) */
    
    if (!is_verb_info(arglist.v.list[3].v.list)) {
	free_var(arglist);
	return make_error_pack(E_TYPE);
    } else if (!retrieve(arglist.v.list[1].v.obj)) {
	free_var(arglist);
	return make_error_pack(E_INVARG);
    } else {
	int		vid;
	Verbdef    *v = match_verbdef(arglist.v.list[2].v.str, 
				      arglist.v.list[1].v.obj, &vid);
	
	if (!v) {
	    free_var(arglist);
	    return make_error_pack(E_VERBNF);
	} else {
	    Objid	new_owner = arglist.v.list[3].v.list[1].v.obj;
	    char   *perm_str = arglist.v.list[3].v.list[2].v.str;
	    int	new_perms = parse_verb_perms(perm_str);
	    char   *new_name = arglist.v.list[3].v.list[3].v.str;
	    
	    if ((!is_wizard(progr) && new_owner != v->owner)
		|| !Verb_perm(progr, v, VP_WRITE)) {
		free_var(arglist);
		return make_error_pack(E_PERM);
	    } else if (new_perms < 0
		       || !retrieve(new_owner)
		       || !is_player(new_owner)) {
		free_var(arglist);
		return make_error_pack(E_INVARG);
	    } else {
		v->owner = new_owner;
		v->perms &= ~VP_PERMMASK;
		v->perms |= new_perms;
		v->name = strdup(new_name);
	    }
	}
    }
    free_var(arglist);
    return no_var_pack();
}

static package
bf_verb_args(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, verb-name) */
    Var r;
    
    if (!retrieve(arglist.v.list[1].v.obj)) {
	free_var(arglist);
	return make_error_pack(E_INVARG);
    } else {
	int		vid;
	Verbdef    *v = match_verbdef(arglist.v.list[2].v.str, 
				      arglist.v.list[1].v.obj, &vid);
	
	if (!v) {
	    free_var(arglist);
	    return make_error_pack(E_INVARG);
	} else if (!Verb_perm(progr, v, VP_READ)) {
	    free_var(arglist);
	    return make_error_pack(E_PERM);
	} else {
	    r = new_list(3);
	    r.v.list[1].type = TYPE_STR;
	    r.v.list[1].v.str =
		strdup(unparse_option((v->perms >> VP_DOBJSHIFT)
				      & VP_OBJMASK));
	    r.v.list[2].type = TYPE_STR;
	    r.v.list[2].v.str = strdup(unparse_prep(v->prep));
	    r.v.list[3].type = TYPE_STR;
	    r.v.list[3].v.str =
		strdup(unparse_option((v->perms >> VP_IOBJSHIFT)
				      & VP_OBJMASK));
	}
    }
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_set_verb_args(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, verb-name, {dobj, prep, iobj}) */
    
    if (!is_args(arglist.v.list[3].v.list)) {
	free_var(arglist);
	return make_error_pack(E_TYPE);
    } else if (!retrieve(arglist.v.list[1].v.obj)) {
	free_var(arglist);
	return make_error_pack(E_INVARG);
    } else {
	int		vid;
	Verbdef    *v = match_verbdef(arglist.v.list[2].v.str, 
				      arglist.v.list[1].v.obj, &vid);
	Var	       *a = arglist.v.list[3].v.list;
	int		dobj = match_option(a[1].v.str);
	int		prep = match_prep(a[2].v.str);
	int		iobj = match_option(a[3].v.str);
	
	if (!v) {
	    free_var(arglist);
	    return make_error_pack(E_VERBNF);
	} else if (dobj == OPT_FAIL
		   || prep == FAILED_MATCH
		   || iobj == OPT_FAIL) {
	    free_var(arglist);
	    return make_error_pack(E_INVARG);
	} else if (!Verb_perm(progr, v, VP_WRITE)) {
	    free_var(arglist);
	    return make_error_pack(E_PERM);
	} else {
	    v->perms = ((v->perms & VP_PERMMASK)
			| (dobj << VP_DOBJSHIFT)
			| (iobj << VP_IOBJSHIFT));
	    v->prep = prep;
	}
    }
    free_var(arglist);
    return no_var_pack();
}

static package
bf_add_verb(Var arglist, Byte next, void *vdata, Objid progr) 
{ /* (object, info, args) */
    
    if (!is_verb_info(arglist.v.list[2].v.list)
	|| !is_args(arglist.v.list[3].v.list)) {
	free_var(arglist);
	return make_error_pack(E_TYPE);
    } else if (!retrieve(arglist.v.list[1].v.obj)) {
	free_var(arglist);
	return make_error_pack(E_INVARG);
    } else {
	Verbdef    *v, *newv;
	Objid	oid = arglist.v.list[1].v.obj;
	Object     *o = retrieve(oid);
	Var	       *info = arglist.v.list[2].v.list;
	Objid	owner = info[1].v.obj;
	int		perms = parse_verb_perms(info[2].v.str);
	char       *names = info[3].v.str;
	Var	       *a = arglist.v.list[3].v.list;
	int		dobj = match_option(a[1].v.str);
	int		prep = match_prep(a[2].v.str);
	int		iobj = match_option(a[3].v.str);
	
	if (perms < 0
	    || !retrieve(owner)
	    || !is_player(owner)
	    || dobj == OPT_FAIL
	    || prep == FAILED_MATCH
	    || iobj == OPT_FAIL) { 
	    free_var(arglist);
	    return make_error_pack(E_INVARG);
	} else if (!Obj_perm(progr, oid, FLAG_WRITE)
		   || (!is_wizard(progr) && progr != owner)) {
	    free_var(arglist);
	    return make_error_pack(E_PERM);
	} else {
	    newv = (Verbdef *) mymalloc(sizeof(Verbdef), M_VERBDEF);
	    newv->name = strdup(names);
	    newv->owner = owner;
	    newv->perms = perms | (dobj << VP_DOBJSHIFT)
		| (iobj << VP_IOBJSHIFT);
	    newv->prep = prep;
	    newv->next = 0;
	    newv->program = 0;
	    if (o->verbdefs) {
		for (v = o->verbdefs; v->next; v = v->next)
		    ;
		v->next = newv;
	    } else
		o->verbdefs = newv;
	}
    }
    free_var(arglist);
    return no_var_pack();
}

static package
bf_delete_verb(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, verb-name) */
    
    if (!retrieve(arglist.v.list[1].v.obj)) {
	free_var(arglist);
	return make_error_pack(E_INVARG);
    } else if (!Obj_perm(progr, arglist.v.list[1].v.obj, FLAG_WRITE)) { 
	free_var(arglist);
	return make_error_pack(E_PERM);
    } else {
	int		vid;
	Object     *o = retrieve(arglist.v.list[1].v.obj);
	Verbdef    *v = match_verbdef(arglist.v.list[2].v.str, 
				      arglist.v.list[1].v.obj, &vid);
	Verbdef    *vv;
	
	if (!v) { 
	    free_var(arglist);
	    return make_error_pack(E_VERBNF);
	} else {
	    vv = o->verbdefs;
	    if (vv == v)
		o->verbdefs = v->next;
	    else {
		while (vv->next != v)
		    vv = vv->next;
		vv->next = v->next;
	    }
	    if (v->program)
		free_program(v->program);
	    if (v->name)
		myfree((void *) v->name, M_STRING);
	    myfree((void *) v, M_VERBDEF);
	}
    }
    free_var(arglist);
    return no_var_pack();
}

static void lister(void *, const char *);

static package
bf_verb_code(Var arglist, Byte next, void *vdata, Objid progr)
{ /* (object, verb-name [, fully-paren [, indent]]) */
    Var r;
    if (!retrieve(arglist.v.list[1].v.obj)) { 
	free_var(arglist);
	return make_error_pack(E_INVARG);
    } else {
	int		vid;
	Verbdef    *v = match_verbdef(arglist.v.list[2].v.str, 
				      arglist.v.list[1].v.obj, &vid);
	
	if (!v) { 
	    free_var(arglist);
	    return make_error_pack(E_VERBNF);
	} else if (!Verb_perm(progr, v, VP_READ)) { 
	    free_var(arglist);
	    return make_error_pack(E_PERM);
	} else {
	    int	parens = 0, indent = 1;
	    
	    if (arglist.v.list[0].v.num >= 3)
		parens = arglist.v.list[3].v.num;
	    if (arglist.v.list[0].v.num >= 4)
		indent = arglist.v.list[4].v.num;
	    
	    r = new_list(0);
	    if (v->program)
		list_prog(v->program, lister, (void *) &r, parens, indent, 
			  MAIN_VECTOR);
	}
    }
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_set_verb_code(Var arglist, Byte next, void *vdata, Objid progr) 
{ /* (object, vname, code) */
    Var r;
    if (!retrieve(arglist.v.list[1].v.obj)) { 
	free_var(arglist);
	return make_error_pack(E_INVARG);
    } else if (!is_code(arglist.v.list[3].v.list)) { 
	free_var(arglist);
	return make_error_pack(E_TYPE);
    } else if (!is_programmer(progr)) { 
	free_var(arglist);
	return make_error_pack(E_PERM);
    } else {
	int		vid;
	Verbdef    *v = match_verbdef(arglist.v.list[2].v.str, 
				      arglist.v.list[1].v.obj, &vid);
	
	if (!v) { 
	    free_var(arglist);
	    return make_error_pack(E_VERBNF);
	} else if (!Verb_perm(progr, v, VP_WRITE)) {
	    free_var(arglist);
	    return make_error_pack(E_PERM);
	} else {
	    struct parser_state	state;
	    Program		       *program;
	    
	    state.code = arglist.v.list[3];
	    state.cur_string = 1;
	    state.cur_char = 0;
	    state.errors = new_list(0);
	    program = parse_program(parser_client, (void *) &state);
	    if (program) {
		if (v->program)
		    free_program(v->program);
		v->program = program; 
	    }
	    r = state.errors;
	}
    }
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_eval(Var arglist, Byte next, void *data, Objid progr)
{
    if (next == 1) {
	
	if (!is_programmer(progr)) {
	    free_var(arglist);
	    return make_error_pack(E_PERM);
	} else {
	    Program	       *program;
	    struct parser_state	state;
	    
	    state.code = arglist;
	    state.cur_string = 1;
	    state.cur_char = 0;
	    state.errors = new_list(0);
	    program = parse_program(parser_client, (void *) &state);
	    free_var(arglist);
	    if (program) {
		free_var(state.errors);
		if (setup_activ_for_eval(program))
		    return make_call_pack(2, NULL);
		else {
		    free_program(program);
		    return make_error_pack(E_MAXREC);
		}
	    } else {
		Var r;
		
		r = new_list(2);
		r.v.list[1].type = TYPE_NUM;
		r.v.list[1].v.num = 0;
		r.v.list[2] = state.errors;
		return make_var_pack(r);
	    }
	}
    } else { /* next == 2 */
	Var r;
	
	r = new_list(2);
	r.v.list[1].type = TYPE_NUM;
	r.v.list[1].v.num = 1;
	r.v.list[2] = arglist; 
	return make_var_pack(r);
    }
}

/* register */

void
register_bf_verb(void)
{ 
    (void) register_function("verbs", 1, 1, bf_verbs, TYPE_OBJ);
    (void) register_function("verb_info", 2, 2, bf_verb_info,
			     TYPE_OBJ, TYPE_STR);
    (void) register_function("set_verb_info", 3, 3, bf_set_verb_info,
			     TYPE_OBJ, TYPE_STR, TYPE_LIST);
    (void) register_function("verb_args", 2, 2, bf_verb_args,
			     TYPE_OBJ, TYPE_STR);
    (void) register_function("set_verb_args", 3, 3, bf_set_verb_args,
			     TYPE_OBJ, TYPE_STR, TYPE_LIST);
    (void) register_function("add_verb", 3, 3, bf_add_verb,
			     TYPE_OBJ, TYPE_LIST, TYPE_LIST);
    (void) register_function("delete_verb", 2, 2, bf_delete_verb,
			     TYPE_OBJ, TYPE_STR);
    (void) register_function("verb_code", 2, 4, bf_verb_code,
			     TYPE_OBJ, TYPE_STR, TYPE_NUM, TYPE_NUM);
    (void) register_function("set_verb_code", 3, 3, bf_set_verb_code,
			     TYPE_OBJ, TYPE_STR, TYPE_LIST);
    (void) register_function("eval", 1, 1, bf_eval, TYPE_STR);
}

/* $log: bf_verb.c,v $
 * Revision 1.2  1992/07/20  23:55:24  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 * */
