char rcsid_bsd_network[] = "$Id: bsd_network.c,v 1.3 1992/08/10 17:22:45 pjames Exp $";

#include "my-types.h"
#include <ctype.h>
#include <errno.h>
#include "my-in.h"
#include "my-ioctl.h"
#include <netdb.h>
#include "my-signal.h"
#include "my-socket.h"
#include "my-stdio.h"
#include "my-stdlib.h"
#include "my-string.h"
#include "my-sys-time.h"
#include "my-unistd.h"

#include "config.h"
#include "exceptions.h"
#include "log.h"
#include "network.h"
#include "server.h"
#include "streams.h"
#include "structures.h"
#include "storage.h"
#include "timers.h"

static int	server_socket;
static int	listening = 0;
static int	pocket_descriptor;	/* fd we keep available in case we need
					 * one and no others are left... */
static const char *server_full_msg =
    "*** Sorry, but the server cannot accept any more connections right now.\r\n*** Please try again later.\r\n";

typedef struct text_block {
    struct text_block  *next;
    unsigned		length;
    char	       *buffer;
    char	       *start;
} text_block;

typedef struct nhandle {
    struct nhandle     *next, **prev;
    server_handle	shandle;
    int			fd;
    char	       *host_name;
    Stream	       *input;
    text_block	       *output_head;
    text_block	      **output_tail;
    unsigned		output_length;
    unsigned		output_lines_flushed;
} nhandle;

static nhandle *all_nhandles = 0;


static void
free_text_block(text_block *b)
{
    myfree(b->buffer, M_NETWORK);
    myfree(b, M_NETWORK);
}

static int
push_output(nhandle *h)
{
    text_block *b;
    int		count;

    if (h->output_lines_flushed > 0) {
	char	buf[100];
	int	length;

	sprintf(buf, "\r\n>> %d line%s flushed <<\r\n",
		h->output_lines_flushed,
		h->output_lines_flushed == 1 ? "" : "s");
	length = strlen(buf);
	count = write(h->fd, buf, length);
	if (count == length)
	    h->output_lines_flushed = 0;
	else
	    return count >= 0  ||  errno == EWOULDBLOCK;
    }
    while ((b = h->output_head) != 0) {
	count = write(h->fd, b->start, b->length);
	if (count < 0)
	    return (errno == EWOULDBLOCK);
	h->output_length -= count;
	if (count == b->length) {
	    h->output_head = b->next;
	    free_text_block(b);
	} else {
	    b->start += count;
	    b->length -= count;
	}
    }
    if (h->output_head == 0)
	h->output_tail = &(h->output_head);
    return 1;
}

static int
pull_input(nhandle *h)
{
    Stream     *s = h->input;
    int		count;
    char	buffer[1024];
    char       *ptr, *end;

    while ((count = read(h->fd, buffer, sizeof(buffer))) > 0) {
	for (ptr = buffer, end = buffer + count; ptr < end; ptr++) {
	    unsigned char	c = *ptr;

	    if (isgraph(c)  ||  c == ' ')
		stream_add_char(s, c);
	    else if (c == '\t')
		stream_add_char(s, ' ');
	    else if (c == '\n')
		server_receive_line(h->shandle, reset_stream(s));
	}
    }
    return (count < 0  &&  errno == EWOULDBLOCK);
}

static const char *
address_string(long a)
{
    static char buf[20];

    sprintf(buf, "%d.%d.%d.%d", (a >> 24) & 0xff, (a >> 16) & 0xff,
	    (a >> 8) & 0xff, a & 0xff);
    return buf;
}

static Exception	ns_timeout;

static void
name_server_timeout(Timer_ID id, Timer_Data data)
{
    RAISE(ns_timeout, 0);
}

static void
make_new_connection(int fd, const char *host_name)
{
    nhandle	       *h;
    network_handle	nh;
    int			yes = 1;

    if (ioctl(fd, FIONBIO, &yes) == -1) {
	log_perror("Setting network connection non-blocking");
	return;
    }
    h = mymalloc(sizeof(nhandle), M_NETWORK);
    if (all_nhandles)
	all_nhandles->prev = &(h->next);
    h->next = all_nhandles;
    h->prev = &all_nhandles;
    all_nhandles = h;
    h->fd = fd;
    h->input = new_stream(100);
    h->output_head = 0;
    h->output_tail = &(h->output_head);
    h->output_length = 0;
    h->output_lines_flushed = 0;
    h->host_name = (char *) mymalloc(strlen(host_name) + 1, M_NETWORK);
    strcpy(h->host_name, host_name);

    nh.ptr = h;
    h->shandle = server_new_connection(nh);
}

static void
get_pocket_descriptor()
{
    pocket_descriptor = dup(0);
    if (!pocket_descriptor) {
	log_perror("Can't get pocket descriptor");
	panic("Need pocket descriptor to continue");
    }
}

static void
accept_new_connection(void)
{
    volatile int	fd;
    struct sockaddr_in	address;
    int			addr_length = sizeof(address);
    struct hostent     *e = 0;
    Timer_ID		id;
    const char	       *host_name;

    fd = accept(server_socket, (struct sockaddr *) &address, &addr_length);
    if (fd < 0) {
	if (errno == EMFILE) {
	    close(pocket_descriptor);
	    fd = accept(server_socket, (struct sockaddr *) &address,
			&addr_length);
	    write(fd, server_full_msg, strlen(server_full_msg));
	    close(pocket_descriptor);
	    get_pocket_descriptor();
	} else if (errno != EINTR)
	    log_perror("Accepting new network connection");
	return;
    }
    id = set_timer(5, name_server_timeout, 0);
    TRY
	e = gethostbyaddr((void *) &address.sin_addr,
			  sizeof(address.sin_addr),
			  AF_INET);
	cancel_timer(id);
    EXCEPT (ns_timeout)
	e = 0;
	reenable_timers();
    ENDTRY;
    host_name = e ? e->h_name : address_string(ntohl(address.sin_addr.s_addr));
    make_new_connection(fd, host_name);
}

static void
close_nhandle(nhandle *h)
{
    text_block *b, *bb;

    (void) push_output(h);
    *(h->prev) = h->next;
    if (h->next)
	h->next->prev = h->prev;
    b = h->output_head;
    while (b) {
	bb = b->next;
	free_text_block(b);
	b = bb;
    }
    free_stream(h->input);
    (void) shutdown(h->fd, 2);
    close(h->fd);
    myfree(h->host_name, M_NETWORK);
    myfree(h, M_NETWORK);
}


/*************************
 * External entry points *
 *************************/

const char *
network_usage_string(void)
{
    return "[port]";
}

int
network_initialize(int argc, char **argv)
{
    int			port;
    char	       *p;
    struct sockaddr_in	address;
    int			option = 1;
    extern int		getdtablesize(void);
    
    if (argc > 1)
	return 0;
    else if (argc == 1) {
	port = strtol(argv[0], &p, 10);
	if (*p != '\0')
	    return 0;
    } else
	port = DEFAULT_PORT;

    get_pocket_descriptor();

    server_socket = socket(AF_INET, SOCK_STREAM, 0);
    if (server_socket < 0) {
	log_perror("Creating server socket");
	return 0;
    }
    if (setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR,
		   (char *) &option, sizeof(option)) < 0) {
	log_perror("Setting server socket options");
	return 0;
    }
    address.sin_family = AF_INET;
    address.sin_addr.s_addr = INADDR_ANY;
    address.sin_port = htons(port);
    if (bind(server_socket, (struct sockaddr *) &address, sizeof(address))) {
	log_perror("Binding server socket");
	close(server_socket);
	return 0;
    }

    /* we don't care about SIGPIPE, we notice it in select() and write() */
    signal(SIGPIPE, SIG_IGN);

    return 1;
}

void
network_send_line(network_handle nh, const char *line)
{
    nhandle    *h = nh.ptr;
    unsigned	line_length = strlen(line);
    unsigned	length = line_length + 2;
    char       *buffer = (char *) mymalloc(length * sizeof(char), M_NETWORK);
    text_block *block = (text_block *) mymalloc(sizeof(text_block), M_NETWORK);

    if (h->output_length + length > MAX_QUEUED_OUTPUT) {
	int		to_flush;
	text_block     *b;
	
	(void) push_output(h);
	to_flush = h->output_length + length - MAX_QUEUED_OUTPUT;
	while (to_flush > 0  &&  (b = h->output_head)) {
	    h->output_length -= b->length;
	    to_flush -= b->length;
	    h->output_lines_flushed++;
	    h->output_head = b->next;
	    free_text_block(b);
	}
	if (h->output_head == 0)
	    h->output_tail = &(h->output_head);
    }
    memcpy(buffer, line, line_length);
    buffer[line_length] = '\r';
    buffer[line_length + 1] = '\n';
    block->buffer = block->start = buffer;
    block->length = length;
    block->next = 0;
    *(h->output_tail) = block;
    h->output_tail = &(block->next);
    h->output_length += length;
}

void
network_process_io(unsigned timeout)
{
    fd_set	input_set, output_set;
    nhandle    *h, *hnext;
    struct timeval	tv;
    int		max_descriptor;
#ifdef sun
    extern void		bzero(char *, int);  /* used in expansion of FD_ZERO */
#endif

    if (!listening) {
	listen(server_socket, 5);
	listening = 1;
    }

    FD_ZERO(&input_set);
    FD_ZERO(&output_set);
    FD_SET(server_socket, &input_set);
    max_descriptor = server_socket;
    for (h = all_nhandles; h; h = h->next) {
	FD_SET(h->fd, &input_set);
	if (h->output_head)
	    FD_SET(h->fd, &output_set);
	if (h->fd > max_descriptor)
	    max_descriptor = h->fd;
    }

    tv.tv_sec = timeout;
    tv.tv_usec = 0;
    /* The casts here are to placate systems that make inconsistent decls of */
    /* select(), like declaring the args to be (int *). */
    if (select(max_descriptor + 1, (void*) &input_set, (void *) &output_set, 0,
	       &tv) < 0) {
	if (errno != EINTR)
	    log_perror("Waiting for network I/O");
	return;
    } else {
	if (FD_ISSET(server_socket, &input_set))
	    accept_new_connection();
	for (h = all_nhandles; h; h = hnext) {
	    hnext = h->next;
	    if ((FD_ISSET(h->fd, &input_set)  &&  !pull_input(h))
		|| (FD_ISSET(h->fd, &output_set)  &&  !push_output(h))) {
		server_close(h->shandle);
		close_nhandle(h);
	    }
	}
    }
}

const char *
network_connection_name(network_handle nh)
{
    nhandle    *h = (nhandle *) nh.ptr;
    static char	buffer[100];

    sprintf(buffer, "%d from %s", h->fd, h->host_name);
    return buffer;
}

#ifdef OUTBOUND_NETWORK
#include "structures.h"

enum error
network_open_connection(Var arglist)
{
    /* These are `static' rather than `volatile' because I can't cope with
     * getting all those nasty little parameter-passing rules right.  This
     * function isn't recursive anyway, so it doesn't matter.
     */
    static char			       *host_name;
    static int				port;
    static struct hostent	       *e;
    static struct sockaddr_in		addr;
    Timer_ID				id;
    int					s;

    if (arglist.v.list[0].v.num != 2)
	return E_ARGS;
    else if (arglist.v.list[1].type != TYPE_STR ||
	     arglist.v.list[2].type != TYPE_NUM)
	return E_TYPE;

    host_name = arglist.v.list[1].v.str;
    port = arglist.v.list[2].v.num;

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    id = set_timer(5, name_server_timeout, 0);
    TRY
	e = gethostbyname(host_name);
	cancel_timer(id);
    EXCEPT (ns_timeout)
	e = 0;
	reenable_timers();
    ENDTRY;
    if (e)
	memcpy(&addr.sin_addr, e->h_addr_list[0], e->h_length);
    else {
	struct in_addr	ia;

	ia.s_addr = inet_addr(host_name);
	if (ia.s_addr == -1)
	    return E_INVARG;
	memcpy(&addr.sin_addr, &ia, sizeof(ia));
    }

    s = socket(AF_INET, SOCK_STREAM, 0);
    if (s < 0) {
	if (errno != ENFILE)
	    log_perror("Making socket in network_open_connection");
	return E_QUOTA;
    }
    if (connect(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
	close(s);
	if (errno == EADDRNOTAVAIL ||
	    errno == ECONNREFUSED  ||
	    errno == ENETUNREACH   ||
	    errno == ETIMEDOUT)
	    return E_INVARG;
	log_perror("Connecting in network_open_connection");
	return E_QUOTA;
    }
    make_new_connection(s, host_name);

    return E_NONE;
}
#endif

void
network_close(network_handle h)
{
    close_nhandle((nhandle *) h.ptr);
}

void
network_shutdown(void)
{
    while (all_nhandles)
	close_nhandle(all_nhandles);
    close(server_socket);
}

/* $Log: bsd_network.c,v $
 * Revision 1.3  1992/08/10  17:22:45  pjames
 * Updated #includes.
 *
 * Revision 1.2  1992/07/20  23:56:16  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 * */
