char rcsid_server[] = "$Id: server.c,v 1.7 1992/08/13 21:28:11 pjames Exp $";

#include "my-types.h"	/* must be first on some systems */
#include "my-signal.h"
#include "my-stdio.h"
#include "my-stdlib.h"
#include "my-unistd.h"
#include <sys/wait.h>

#include "config.h"
#include "functions.h"
#include "list.h"
#include "log.h"
#include "network.h"
#include "parse_command.h"
#include "random.h"
#include "server.h"
#include "storage.h"
#include "streams.h"
#include "structures.h"
#include "tasks.h"
#include "timers.h"
#include "utils.h"
#include "version.h"

#include "execute.h"

static pid_t		parent_pid;
static int		in_child = 0;

static char	       *dump_db_name;
static int		dump_generation = 0;

static const char      *shutdown_message = 0;	/* shut down if non-zero */

typedef enum {CHKPT_OFF, CHKPT_TIMER, CHKPT_SIGNAL} Checkpoint_Reason;
static Checkpoint_Reason checkpoint_requested = CHKPT_OFF;

typedef enum { DUMP_SHUTDOWN, DUMP_CHECKPOINT, DUMP_PANIC } Dump_Reason;
const char *reason_names[] = { "DUMPING", "CHECKPOINTING", "PANIC-DUMPING" };

typedef struct shandle {
    struct shandle     *next, **prev;
    network_handle	nhandle;
    time_t		connection_time;
    time_t		last_activity_time;
    Objid		player;
    task_queue		tasks;
    int			disconnect_me;
} shandle;

shandle	*active_shandles = 0;


static void
free_shandle(shandle *h)
{
    *(h->prev) = h->next;
    if (h->next)
	h->next->prev = h->prev;

    free_task_queue(h->tasks);

    myfree(h, M_NETWORK);
}

static void
dump_database(Dump_Reason reason)
{
    Stream     *s = new_stream(100);
    char       *temp_name;
    FILE       *f;

  retryDumping:

    stream_printf(s, "%s.#%d#", dump_db_name, dump_generation);
    remove(reset_stream(s));	/* Remove previous checkpoint */

    if (reason == DUMP_PANIC)
	stream_printf(s, "%s.PANIC", dump_db_name);
    else {
	dump_generation++;
	stream_printf(s, "%s.#%d#", dump_db_name, dump_generation);
    }
    temp_name = reset_stream(s);

    log("%s on %s ...\n", reason_names[reason], temp_name);

    if (reason == DUMP_CHECKPOINT) {
	pid_t	result = fork();

	if (result == 0)
	    in_child = 1;
	else {			/* in parent, possibly with an error */
	    if (result < 0)
		log_perror("Forking checkpointer");
	    else
		reset_command_history();
	    free_stream(s);
	    return;
	}
    }

    if ((f = fopen(temp_name, "w")) != NULL) {
	if (!db_write(f)) {
	    log_perror("Trying to dump database");
	    fclose(f);
	    remove(temp_name);
	    if (reason == DUMP_CHECKPOINT) {
		errlog("Abandoning checkpoint attempt...\n");
		exit(0);
	    } else {
		errlog("Waiting %d seconds and retrying dump...\n",
		       DUMP_RETRY_INTERVAL);
		timer_sleep(DUMP_RETRY_INTERVAL);
		goto retryDumping;
	    }
	}
	fclose(f);
	log("%s on %s finished\n", reason_names[reason], temp_name);
	if (reason != DUMP_PANIC  &&  rename(temp_name, dump_db_name) < 0)
	    log_perror("Renaming temporary dump file");
    } else
	log_perror("Opening temporary dump file");

    if (reason == DUMP_CHECKPOINT)
	/* We're a child, so we'd better go away. */
	exit(0);
    else
	free_stream(s);
}

static void
send_shutdown_message(const char *msg)
{
    shandle	*h;
    Stream	*s = new_stream(100);
    char	*message;

    stream_printf(s, "*** Shutting down: %s ***", msg);
    message = stream_contents(s);
    for (h = active_shandles; h; h = h->next)
	network_send_line(h->nhandle, message);
    free_stream(s);
}

static void
abort_server(void)
{
    signal(SIGINT,  SIG_DFL);
    signal(SIGTERM, SIG_DFL);
    signal(SIGFPE,  SIG_DFL);
    signal(SIGHUP,  SIG_DFL);
    signal(SIGILL,  SIG_DFL);
    signal(SIGQUIT, SIG_DFL);
    signal(SIGSEGV, SIG_DFL);
    signal(SIGBUS,  SIG_DFL);
    signal(SIGUSR1, SIG_DFL);
    signal(SIGUSR2, SIG_DFL);
    signal(SIGCHLD, SIG_DFL);

    abort();
}

void
panic(const char *message)
{
    static int	in_panic = 0;

    errlog("PANIC%s: %s\n", in_child ? " (in child)" : "", message);
    if (in_panic) {
	errlog("RECURSIVE PANIC: aborting\n");
	abort_server();
    }
    in_panic = 1;

    log_command_history();

    if (in_child) {		/* We're a forked checkpointer */
	errlog("Child shutting down parent via USR1 signal\n");
	kill(parent_pid, SIGUSR1);
	_exit(1);
    }

    send_shutdown_message("server panic");
    network_shutdown();
    dump_database(DUMP_PANIC);

    abort_server();
}

static void
panic_signal(int sig)
{
    char message[100];
    
    sprintf(message, "Caught signal %d", sig);
    panic(message);
}

static void
shutdown_signal(int sig)
{
    shutdown_message = "shutdown signal received";
}

static void
checkpoint_signal(int sig)
{
    checkpoint_requested = CHKPT_SIGNAL;

    signal(sig, checkpoint_signal);
}

static void
child_completed_signal(int sig)
{
#ifdef USE_POSIX
    while (waitpid(-1, 0, WNOHANG) > 0)
	;
#else
#ifdef MIPS_SVR3
    while (wait2(0, WNOHANG) >= 0)
	;
#else  /* BSD */
    while (wait3(0, WNOHANG, 0) >= 0)
	;
#endif
#endif

    signal(sig, child_completed_signal);
}

static void
setup_signals(void)
{
    signal(SIGINT,  panic_signal);
    signal(SIGTERM, panic_signal);
    signal(SIGFPE,  panic_signal);
    signal(SIGHUP,  panic_signal);
    signal(SIGILL,  panic_signal);
    signal(SIGQUIT, panic_signal);
    signal(SIGSEGV, panic_signal);
    signal(SIGBUS,  panic_signal);

    signal(SIGUSR1, shutdown_signal);   /* remote shutdown signal */
    signal(SIGUSR2, checkpoint_signal); /* remote checkpoint signal */

    signal(SIGCHLD, child_completed_signal);
}

static void
checkpoint_timer(Timer_ID id, Timer_Data data)
{
    checkpoint_requested = CHKPT_TIMER;
}

static void
set_checkpoint_timer(int first_time)
{
    Var			v;
    int			interval;
    static Timer_ID	last_checkpoint_timer;

    v = get_system_property("dump_interval");
    if (v.type != TYPE_NUM  ||  v.v.num < 60)
	interval = DEFAULT_DUMP_INTERVAL;
    else
	interval = v.v.num;

    if (!first_time)
	cancel_timer(last_checkpoint_timer);
    last_checkpoint_timer = set_timer(interval, checkpoint_timer, 0);
}

static const char *
object_name(Objid oid)
{
    static Stream      *s = 0;
    Object	       *o = retrieve(oid);

    if (!s)
	s = new_stream(30);

    if (o)
	stream_printf(s, "%s (#%d)", o->name, oid);
    else
	stream_printf(s, "#%d", oid);

    return reset_stream(s);
}

static void
run_confuncs(Objid player)
{
    if (retrieve(player)) {
	Objid	location = retrieve(player)->location;

	if (location != NOTHING) {
	    Var	args;

	    args = new_list(1);
	    args.v.list[1].type = TYPE_OBJ;
	    args.v.list[1].v.obj = player;
	    run_server_task(player, location, "confunc", args, 0);
	}
	run_server_task(player, player, "confunc", new_list(0), 0);
    }
}

static void
run_disfuncs(Objid player)
{
    if (retrieve(player)) {
	Objid	location = retrieve(player)->location;

	run_server_task(player, player, "disfunc", new_list(0), 0);

	if (location != NOTHING) {
	    Var	args;

	    args = new_list(1);
	    args.v.list[1].type = TYPE_OBJ;
	    args.v.list[1].v.obj = player;
	    run_server_task(player, location, "disfunc", args, 0);
	}
    }
}

static void
main_loop(void)
{
    run_server_task(-1, SYSTEM_OBJECT, "server_started", new_list(0), 0);
    set_checkpoint_timer(1);

    while (shutdown_message == 0) {
	int		task_seconds = next_task_start();
	unsigned	max_seconds = 5;
	unsigned	timeout = (task_seconds < 0
				       ||  task_seconds > max_seconds)
				  ? max_seconds
				  : task_seconds;

	network_process_io(timeout);
	
	run_ready_tasks();
	    
	if (checkpoint_requested != CHKPT_OFF) {
	    if (checkpoint_requested == CHKPT_SIGNAL)
		log("CHECKPOINTING due to remote request signal.\n");
	    checkpoint_requested = CHKPT_OFF;
	    dump_database(DUMP_CHECKPOINT);
	    set_checkpoint_timer(0);
	}

	{			/* Get rid of old un-logged-in connections */
	    time_t	now = time(0);
	    shandle    *h, *nexth;

	    for (h = active_shandles; h; h = nexth) {
		nexth = h->next;
		if (h->connection_time == 0
		        &&  now - h->last_activity_time > CONNECT_TIMEOUT) {
		    log("TIMEOUT: #%d on %s\n",
			h->player,
			network_connection_name(h->nhandle));
		    network_send_line(h->nhandle,
				      "*** Timed-out waiting for login. ***");
		    network_close(h->nhandle);
		    free_shandle(h);
		} else if (h->connection_time != 0 && !retrieve(h->player)) {
		    log("RECYCLED: #%d on %s\n",
			h->player,
			network_connection_name(h->nhandle));
		    network_send_line(h->nhandle, "*** Recycled ***");
		    network_close(h->nhandle);
		    free_shandle(h);
		} else if (h->disconnect_me) {
		    run_disfuncs(h->player);
		    log("DISCONNECTED: %s on %s\n",
			object_name(h->player),
			network_connection_name(h->nhandle));
		    network_send_line(h->nhandle, "*** Disconnected ***");
		    network_close(h->nhandle);
		    free_shandle(h);
		}
	    }
	}

    }

    log("SHUTDOWN: %s\n", shutdown_message);
    send_shutdown_message(shutdown_message);
}

static shandle *
find_shandle(Objid player)
{
    shandle    *h;

    for (h = active_shandles; h; h = h->next)
	if (h->player == player)
	    return h;

    return 0;
}


/*
 * Exported interface
 */

static Objid 	next_unconnected_player = NOTHING - 1;

server_handle
server_new_connection(network_handle nh)
{
    shandle    	       *h = (shandle *) mymalloc(sizeof(shandle), M_NETWORK);
    server_handle	result;

    h->next = active_shandles;
    h->prev = &active_shandles;
    if (active_shandles)
	active_shandles->prev = &(h->next);
    active_shandles = h;

    h->nhandle = nh;
    h->connection_time = 0;
    h->last_activity_time = time(0);
    h->player = next_unconnected_player--;
    h->tasks = new_task_queue(h->player);
    h->disconnect_me = 0;

    enqueue_input_task(h->tasks, "");

    log("ACCEPT: #%d on %s\n", h->player, network_connection_name(nh));

    result.ptr = h;
    return result;
}

void
server_receive_line(server_handle sh, const char *line)
{
    shandle    *h = (shandle *) sh.ptr;

    h->last_activity_time = time(0);
    enqueue_input_task(h->tasks, line);
}

void
server_close(server_handle sh)
{
    shandle    *h = (shandle *) sh.ptr;
    Objid	player = h->player;

    log("CLIENT DISCONNECTED: %s on %s\n",
	object_name(player),
	network_connection_name(h->nhandle));
    free_shandle(h);
    run_disfuncs(player);
}

void
player_connected(Objid old_id, Objid new_id, int is_newly_created)
{
    shandle    *existing_h = find_shandle(new_id);
    shandle    *new_h = find_shandle(old_id);

    if (!new_h)
	panic("Non-existent shandle connected");
    
    new_h->player = new_id;
    new_h->connection_time = time(0);
    
    if (existing_h) {
	/* network_connection_name is allowed to reuse the same string
	 * storage, so we have to copy one of them.
	 */
	char   *name1 = strdup(network_connection_name(existing_h->nhandle));

	log("REDIRECTED: %s, was %s, now %s\n",
	    object_name(new_id),
	    name1,
	    network_connection_name(new_h->nhandle));
	myfree(name1, M_STRING);
	network_send_line(existing_h->nhandle,
			  "*** Redirecting connection to new port ***");
	network_send_line(new_h->nhandle,
			  "*** Redirecting old connection to this port ***");
	network_close(existing_h->nhandle);
	free_shandle(existing_h);
    } else {
	log("%s: %s on %s\n",
	    is_newly_created ? "CREATED" : "CONNECTED",
	    object_name(new_h->player),
	    network_connection_name(new_h->nhandle));
	network_send_line(new_h->nhandle,
			  is_newly_created ? "*** Created ***"
					   : "*** Connected ***");
	run_confuncs(new_h->player);
    }
}

void
notify(Objid player, const char *message)
{
    shandle    *h = find_shandle(player);

    if (h)
	network_send_line(h->nhandle, message);
}

Var
connected_players(void)
{
    shandle    *h;
    int		count = 0;
    Var		result;

    for (h = active_shandles; h; h = h->next)
	if (h->connection_time != 0)
	    count++;

    result = new_list(count);
    count = 0;

    for (h = active_shandles; h; h = h->next) {
	if (h->connection_time != 0) {
	    count++;
	    result.v.list[count].type = TYPE_OBJ;
	    result.v.list[count].v.obj = h->player;
	}
    }

    return result;
}

int
idle_seconds(Objid player)
{
    shandle    *h = find_shandle(player);

    if (h)
	return time(0) - h->last_activity_time;
    else
	return -1;
}

int
connected_seconds(Objid player)
{
    shandle    *h = find_shandle(player);

    if (h  &&  h->connection_time != 0)
	return time(0) - h->connection_time;
    else
	return -1;
}

void
boot_player(Objid player)
{
    shandle    *h = find_shandle(player);

    if (h)
	h->disconnect_me = 1;
}

int
main(int argc, char **argv)
{
    FILE       *f;
    char       *input_db = argv[1],
	       *dump_db = argv[2];

    if (argc < 3  ||  !network_initialize(argc - 3, argv + 3)) {
	errlog("Usage: %s input-db dump-db %s\n",
	       argv[0],
	       network_usage_string());
	exit(1);
    }

    register_bi_functions();

    log("STARTING: Version %s of the LambdaMOO server\n", server_version);
    log("          (Task timeouts measured in %s seconds.)\n",
	virtual_timer_available() ? "server CPU" : "wall-clock");

    if ((f = fopen(input_db, "r")) == NULL) {
	errlog("Cannot open database file: %s\n", input_db);
	exit(1);
    }

    log("LOADING: %s\n", input_db);

    if (db_read(f) < 0) {
	errlog("Cannot parse database file!\n");
	exit(1);
    }

    log("LOADING: %s done, will dump new database on %s\n", input_db, dump_db);

    fclose(f);

    SRANDOM(time(0));

    dump_db_name = dump_db;

    parent_pid = getpid();
    setup_signals();
    reset_command_history();
    build_preptable();
    
    main_loop();

    network_shutdown();
    dump_database(DUMP_SHUTDOWN);

    return 0;
}

void
request_shutdown(Objid player, const char *message)
{
    /*
     * The stream 's' and its contents will leak, but we're shutting down,
     * so it doesn't really matter.
     */

    Stream     *s = new_stream(100);

    stream_printf(s, "shutdown() called by %s: %s",
		  object_name(player),
		  message);
    shutdown_message = stream_contents(s);
}

void
request_checkpoint(void)
{
    checkpoint_requested = 1;
}

const char *
player_connection_name(Objid player)
{
    shandle    *h = find_shandle(player);

    if (h)
	return network_connection_name(h->nhandle);
    else
	return 0;
}

#ifdef OUTBOUND_NETWORK
Var
open_network_connection(Var arglist)
{
    enum error	e = network_open_connection(arglist);
    Var		r;

    if (e == E_NONE) {
	/* The connection was successfully opened, implying that
	 * server_new_connection was called, implying and a new negative
	 * player number was allocated for the connection.  Thus, the old
	 * value of next_unconnected_player is the number of our connection.
	 */
	r.type = TYPE_OBJ;
	r.v.obj = next_unconnected_player + 1;
    } else {
	r.type = TYPE_ERR;
	r.v.err = e;
    }

    return r;
}
#endif

/* $Log: server.c,v $
 * Revision 1.7  1992/08/13  21:28:11  pjames
 * Now only calls one procedure to register built in functions.
 *
 * Revision 1.6  1992/08/12  01:50:42  pjames
 * run_ready_tasks() was moved to before logoff actions, so that logoffs
 * would happen faster.
 *
 * Revision 1.5  1992/08/10  16:55:31  pjames
 * Updated #includes.  Added register_bf_log().
 *
 * Revision 1.4  1992/07/30  23:22:26  pavel
 * Added a remote checkpoint-request signal (SIGUSR2), analogous to the existing
 * shutdown-requestion signal (SIGUSR1).
 *
 * Revision 1.3  1992/07/30  00:40:59  pavel
 * Add support for compiling on RISC/os 4.52 and NonStop-UX A22.
 *
 * Revision 1.2  1992/07/21  00:06:27  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 * */
