char rcsid_tasks[] = "$Id: tasks.c,v 1.5 1992/08/12 01:51:16 pjames Exp $";

#include "my-string.h"
#include "my-time.h"

#include "config.h"
#include "list.h"
#include "log.h"
#include "match.h"
#include "parse_command.h"
#include "random.h"
#include "server.h"
#include "storage.h"
#include "streams.h"
#include "structures.h"
#include "tasks.h"
#include "utils.h"
#include "execute.h"
#include "eval_vm.h"
#include "unparse.h"
#include "parser.h"

typedef struct forked_task {
    int		id;
    Program    *program;
    activation  a;
    Var        *rt_env;
    int		f_index;
    time_t	start_time;
} forked_task;

typedef struct suspended_task {
    vm 	   the_vm;
    time_t start_time; 
    int    id;
} suspended_task;

/* moved to tasks.h
typedef enum { TASK_INPUT, TASK_FORKED, TASK_SUSPENDED } task_kind;
*/

typedef struct task {
    struct task	       	       	       *next;
    task_kind			       	kind;
    union {
	char	       *input;
	forked_task	forked;
	suspended_task	suspended;
    }					t;
} task;

typedef struct tqueue {
    /*
     * A task queue can be in one of four possible states, depending upon the
     * values of its `player' and `connected' slots.
     *
     * When people first connect to the MOO, they are assigned negative player
     * object numbers.  Once they log in (either by creating a new player
     * object or by connecting to an existing one), the connection is changed
     * to refer to the appropriate, non-negative, player number.
     *
     * Input tasks for logged-in players (with non-negative `player' slots) are
     * interpreted as normal MOO commands.  Those for anonymous players (with
     * negative `player' slots) are treated specially: they are passed to a
     * particular verb on the system object.  If that verb returns a valid
     * player object number, then that is used as the new player number for the
     * connection.
     *
     * The `connected' field is true iff this queue has an associated open
     * network connection.  Queues without such connections may nonetheless
     * contain input tasks, presumably typed ahead before disconnection.  Input
     * tasks in unconnected queues are discarded.
     *
     * If an unconnected queue becomes empty, it is destroyed.
     */
    struct tqueue      *next, **prev;	/* prev only valid on empty_tqueues */
    Objid		player;
    int			connected;
    task	       *first, **last;
    int			usage;		/* a kind of inverted priority */
    char	       *output_prefix, *output_suffix;
    Stream	       *program_stream;
    Objid		program_object;
    int			program_verbnum;

    char 		reading; /* boolean */
    vm			reading_vm;
    int 		reading_task_id;
} tqueue;

#define NO_USAGE	-1

static int	current_task_id;
static tqueue  *empty_tqueues = NULL, *active_tqueues = NULL;
static task    *waiting_tasks = NULL;  /* forked and suspended tasks */

#define GET_START_TIME(ttt) \
    (ttt->kind == TASK_FORKED \
     ? ttt->t.forked.start_time \
     : ttt->t.suspended.start_time)



static void
deactivate_tqueue(tqueue *tq)
{
    tq->usage = NO_USAGE;

    tq->next = empty_tqueues;
    tq->prev = &empty_tqueues;
    if (empty_tqueues)
	empty_tqueues->prev = &(tq->next);
    empty_tqueues = tq;
}

static void
activate_tqueue(tqueue *tq)
{
    tqueue    **qq = &active_tqueues;
    
    while (*qq  &&  (*qq)->usage <= tq->usage)
	qq = &((*qq)->next);
    
    tq->next = *qq;
    tq->prev = 0;
    *qq = tq;
}

static void
ensure_usage(tqueue *tq)
{
    if (tq->usage == NO_USAGE) {
	tq->usage = active_tqueues ? active_tqueues->usage : 0;

	/* Remove tq from empty_tqueues... */
	*(tq->prev) = tq->next;
	if (tq->next)
	    tq->next->prev = tq->prev;

	/* ...and put it on active_tqueues. */
	activate_tqueue(tq);
    }
}

static tqueue *
find_tqueue(Objid player, int create_if_not_found)
{
    tqueue     *tq;
    
    for (tq = active_tqueues; tq; tq = tq->next)
	if (tq->player == player)
	    return tq;

    for (tq = empty_tqueues; tq; tq = tq->next)
	if (tq->player == player)
	    return tq;

    if (!create_if_not_found)
	return 0;
    
    tq = (tqueue *) mymalloc(sizeof(tqueue), M_TASK);

    deactivate_tqueue(tq);

    tq->player = player;
    tq->connected = 0;

    tq->first = 0;
    tq->last = &(tq->first);

    tq->output_prefix = tq->output_suffix = 0;
    tq->program_stream = 0;

    tq->reading = 0;

    return tq;
}

static void
free_tqueue(tqueue *tq)
{
    /* Precondition: tq is on empty_tqueues */

    if (tq->output_prefix)
	myfree(tq->output_prefix, M_STRING);
    if (tq->output_suffix)
	myfree(tq->output_suffix, M_STRING);
    if (tq->program_stream)
	free_stream(tq->program_stream);
    if (tq->reading)
	free_vm(tq->reading_vm, 1);

    *(tq->prev) = tq->next;
    if (tq->next)
	tq->next->prev = tq->prev;

    myfree(tq, M_TASK);
}

static void
enqueue_task(tqueue *tq, task *t)
{
    *(tq->last) = t;
    tq->last = &(t->next);
    t->next = 0;
}

static task *
dequeue_task(tqueue *tq)
{
    task       *t = tq->first;

    if (t) {
	tq->first = t->next;
	if (t->next == 0)
	    tq->last = &(tq->first);
	else
	    t->next = 0;
    }

    return t;
}

static void
free_task(task *t, int strong)
{	/* for FORKED tasks, strong == 1 means free the rt_env also.
	   for SUSPENDED tasks, strong == 1 means free the vm also. */
    switch (t->kind) {
      case TASK_INPUT:
	myfree(t->t.input, M_STRING);
	break;
      case TASK_FORKED:
	if (strong)
	    free_rt_env(t->t.forked.rt_env, 
			t->t.forked.program->num_var_names);
	free_program(t->t.forked.program);
	break;
      case TASK_SUSPENDED:
	if (strong)
	    free_vm(t->t.suspended.the_vm, 1);	    
	break;
    }
    myfree(t, M_TASK);
}

static int
new_task_id(void)
{
    return RANDOM();
}

static void
start_programming(tqueue *tq, char *argstr)
{
    Objid	player = tq->player;
    Objid	oid;
    int		verbnum;
    Verbdef    *v = match_verbref(player, argstr, &oid, &verbnum);
    
    if (oid < 0)
	object_match_failed(player, oid, argstr);
    else if (verbnum == NOTHING)
	notify(player, "You must specify a verb; use the format object:verb.");
    else if (v == 0)
	notify(player, "That object does not have that verb definition.");
    else if (!Verb_perm(player, v, VP_WRITE))
	notify(player, "Permission denied.");
    else {
	Stream *str = new_stream(100);
	
	stream_printf(str, "Now programming %s:%s(%d).  Use \".\" to end.",
		      retrieve(oid)->name, v->name, verbnum);
	notify(player, reset_stream(str));
	tq->program_stream = str;
	tq->program_object = oid;
	tq->program_verbnum = verbnum;
    }
}

struct state {
    Objid	player;
    int		nerrors;
    char       *input;
};

static void
my_error(void *data, const char *msg)
{
    struct state       *s = data;

    notify(s->player, msg);
    s->nerrors++;
}

static void
my_ungetc(void *data, char c)
{
    struct state       *s = data;

    s->input--;
}

static int
my_getc(void *data)
{
    struct state       *s = data;

    if (*(s->input) != '\0')
	return *(s->input++);
    else
	return EOF;
}

static Parser_Client	client = { my_error, my_ungetc, my_getc };

static void
end_programming(tqueue *tq)
{
    Objid	player = tq->player;

    if (!retrieve(tq->program_object))
	notify(player, "That object appears to have disappeared ...");
    else {
	Verbdef	       *v = retrieve(tq->program_object)->verbdefs;
	int		i = tq->program_verbnum;

	while (v  &&  i > 0) {
	    i--;
	    v = v->next;
	}
	
	if (!v)
	    notify(player, "That verb appears to have disappeared ...");
	else {
	    struct state s;
	    Program    *program;
	    char	buf[30];

	    s.player = tq->player;
	    s.nerrors = 0;
	    s.input = stream_contents(tq->program_stream);
	    
	    program = parse_program(client, (void *) &s);
	    
	    sprintf(buf, "%d error(s).", s.nerrors);
	    notify(player, buf);
	    
	    if (program) {
		if (v->program)
		    free_program(v->program);
		v->program = program;
		notify(player, "Verb programmed.");
	    } else
		notify(player, "Verb not programmed.");
	}
    }

    free_stream(tq->program_stream);
    tq->program_stream = 0;
}

static void
set_delimiter(char **slot, char *string)
{
    if (*slot)
	myfree(*slot, M_STRING);
    if (*string == '\0')
	*slot = 0;
    else
	*slot = strdup(string);
}

static int
do_command_task(tqueue *tq, char *command)
{
    if (tq->program_stream) {		/* We're programming */
	if (strcmp(command, ".") == 0)  /* Done programming */
	    end_programming(tq);
	else
	    stream_printf(tq->program_stream, "%s\n", command);
    } else {
	Parse_Info	pi;
	Program	       *p;
	int		junk;
	Objid		location = retrieve(tq->player)->location;

	init_pi(&pi);
	pi.next = 0;
	pi.player = pi.caller = tq->player;

	if (!parse_command(command, &pi))
	    return 0;

	if (is_programmer(pi.player)  &&  verbcasecmp(".pr*ogram", pi.verb)) {
	    if (pi.args.v.list[0].v.num != 1)
		notify(pi.player, "Usage:  .program object:verb");
	    else
		start_programming(tq, pi.args.v.list[1].v.str);
	} else if (strcmp(pi.verb, "PREFIX") == 0)
	    set_delimiter(&(tq->output_prefix), pi.argstr);
	else if (strcmp(pi.verb, "SUFFIX") == 0)
	    set_delimiter(&(tq->output_suffix), pi.argstr);
	else {
	    junk = ((p = find_verb_on(pi.player, pi.player, pi.verb, &pi, 1))
		    || (p = find_verb_on(location, location, pi.verb, &pi, 1))
		    || (p = find_verb_on(pi.dobj, pi.dobj, pi.verb, &pi, 1))
		    || (p = find_verb_on(pi.iobj, pi.iobj, pi.verb, &pi, 1))
		    || (p = find_verb_on(location, location, "huh", &pi, 0)));
	
	    if (tq->output_prefix)
		notify(pi.player, tq->output_prefix);
	    if (p) {
		current_task_id = new_task_id();
		do_input_task(p, &pi);
	    } else
		notify(pi.player, "I couldn't understand that.");
	    if (tq->output_suffix)
		notify(pi.player, tq->output_suffix);
	}

	free_pi(&pi, 0);
    }

    return 1;
}

static int
do_login_task(tqueue *tq, char *command)
{
    Var		result;
    Var		args;
    Objid	old_max_object = db_size() - 1;

    result.type = TYPE_NUM;	/* In case the DB does not contain an
				 * implementation of #0:do_login_command
				 */

    args = parse_into_wordlist(command);
    run_server_task(tq->player, SYSTEM_OBJECT,
		    "do_login_command", args, &result);
    if (result.type == TYPE_OBJ  &&  is_player(result.v.obj)) {
	Objid	new_player = result.v.obj;
	Objid	old_player = tq->player;
	tqueue *old_tq = find_tqueue(new_player, 0);
	task   *t;
	
	if (old_tq) {		/* Copy over tasks from old queue for player */
	    while ((t = dequeue_task(old_tq)) != 0) {
		if (t->kind == TASK_INPUT)
		    free_task(t, 0);
		else /* FORKED or SUSPENDED */
		    enqueue_task(tq, t);
	    }
	    old_tq->player = NOTHING; /* it'll be freed by run_ready_tasks */
	}
	tq->player = new_player;

	player_connected(old_player, new_player, new_player > old_max_object);
    }

    free_var(result);
    return 1;
}

/*
 * Exported interface
 */

task_queue
new_task_queue(Objid player)
{
    task_queue	result;
    tqueue     *tq = result.ptr = find_tqueue(player, 1);

    tq->connected = 1;

    return result;
}

void
free_task_queue(task_queue q)
{
    tqueue     *tq = (tqueue *) q.ptr;

    tq->connected = 0;
}

void
enqueue_input_task(task_queue q, const char *input)
{
    tqueue     *tq = (tqueue *) q.ptr;
    task       *t = (task *) mymalloc(sizeof(task), M_TASK);

    t->kind = TASK_INPUT;
    t->t.input = strdup(input);

    ensure_usage(tq);

    /* Input tasks go before all non-input tasks. */
    {
	task  **tt = &(tq->first);

	while (*tt  &&  (*tt)->kind == TASK_INPUT)
	    tt = &((*tt)->next);

	t->next = *tt;
	*tt = t;
	if (!t->next)
	    tq->last = &(t->next);
    }
}

static void
enqueue_waiting(task *t)
{ /* either FORKED or SUSPENDED */

    time_t 	start_time = GET_START_TIME(t);

    if (!waiting_tasks  ||  start_time < GET_START_TIME(waiting_tasks)) {
	t->next = waiting_tasks;
	waiting_tasks = t;
    } else {
	task   *tt;
	
	for (tt = waiting_tasks; tt->next; tt = tt->next)
	    if (start_time < GET_START_TIME(tt->next))
		break;
	t->next = tt->next;
	tt->next = t;
    }	
}	

static void
enqueue_ft(Program *program, activation a, Var *rt_env,
	   unsigned f_index, time_t start_time, int id)
{
    task       *t = (task *) mymalloc(sizeof(task), M_TASK);

    t->kind = TASK_FORKED;
    t->t.forked.program = program;
    t->t.forked.a = a;
    t->t.forked.a.verb = strdup(t->t.forked.a.verb);
    t->t.forked.a.verbname = strdup(t->t.forked.a.verbname);
    t->t.forked.rt_env = rt_env;
    t->t.forked.f_index = f_index;
    t->t.forked.start_time = start_time;
    t->t.forked.id = id;
    
    enqueue_waiting(t);
}

int
enqueue_forked_task(Program *program, activation a, Var *rt_env, 
		    unsigned f_index, unsigned after_seconds)
{
    int		id = new_task_id();
    
    enqueue_ft(program, a, rt_env, f_index, time(0) + after_seconds, id);
    return id;
}

void
enqueue_suspended_task(vm the_vm, unsigned after_seconds)
{
    task       *t = (task *) mymalloc(sizeof(task), M_TASK);

    t->kind = TASK_SUSPENDED;
    t->t.suspended.the_vm = the_vm;
    t->t.suspended.start_time = time(0) + after_seconds;
    t->t.suspended.id = current_task_id; /* remember old id */
    
    enqueue_waiting(t);
}

enum error
make_reading_task(vm the_vm, Var opt_player)   
{
    Objid	player = opt_player.type == TYPE_OBJ
					 ? opt_player.v.obj
					 : inputting_player(the_vm);
    tqueue     *tq = find_tqueue(player, 0);
    
    if (!tq  ||  tq->reading  ||
	(!tq->connected && (!tq->first  ||  tq->first->kind != TASK_INPUT)))
	return E_INVARG;
    else {
	tq->reading = 1;
	tq->reading_vm = the_vm;
	tq->reading_task_id = current_task_id;
	return E_NONE;
    }
}

int
next_task_start(void)
{
    tqueue     *tq;

    for (tq = active_tqueues; tq; tq = tq->next)
	if (tq->first != 0)
	    return 0;

    if (waiting_tasks != 0) {
	int wait = (waiting_tasks->kind == TASK_FORKED
		    ? waiting_tasks->t.forked.start_time 
		    : waiting_tasks->t.suspended.start_time) - time(0);
	return (wait >= 0) ? wait : 0;
    } 

    else return -1;
}  

void
run_ready_tasks(void)
{
    task       *t, *next_t;
    time_t	now = time(0);
    tqueue     *tq, *next_tq;

    for (t = waiting_tasks; t && GET_START_TIME(t) <= now; t = next_t) {
	Objid progr = (t->kind == TASK_FORKED
		       ? t->t.forked.a.progr
		       : progr_of_cur_verb(t->t.suspended.the_vm)); 
	tqueue	*tq = find_tqueue(progr, 1);

	next_t = t->next;
	ensure_usage(tq);
	enqueue_task(tq, t);
    }
    waiting_tasks = t;
    
    {
	int	did_one = 0;
	time_t	start = time(0);
	
	while (active_tqueues  &&  !did_one) {
	    /* Loop over tqueues, looking for a task */
	    tq = active_tqueues;
	    
	    while (!did_one  &&  (t = dequeue_task(tq)) != 0) {
		/* Loop over tasks, looking for runnable one */
		switch (t->kind) {
		  case TASK_INPUT:
		    if (tq->reading) {
			tq->reading = 0; 
			current_task_id = tq->reading_task_id;
			resume_reading_vm(tq->reading_vm, t->t.input);
			did_one = 1;
		    } else if (tq->connected) {
			add_command_to_history(tq->player, t->t.input);
			did_one = (tq->player >= 0
				   ? do_command_task
				   : do_login_task)(tq, t->t.input);
			}
		    break;
		  case TASK_FORKED:
		    {
			forked_task 	ft;
			
			ft = t->t.forked;
			current_task_id = ft.id;
			do_forked_task(ft.program, ft.rt_env, ft.a,
				       ft.f_index);
			did_one = 1;
		    }
		    break; 
		  case TASK_SUSPENDED:
		    current_task_id = t->t.suspended.id;
		    resume_suspended_task(t->t.suspended.the_vm);
		    did_one = 1;
		    break;
		}
		free_task(t, 0);
	    }

	    active_tqueues = tq->next;

	    if (did_one) {
		/* Bump the usage level of this tqueue */
		time_t	end = time(0);
		
		tq->usage += end - start;
		activate_tqueue(tq);
	    } else {
		/* There was nothing on this tqueue, so deactivate it */
		deactivate_tqueue(tq);
	    }
	}
    }

    /* Free any unconnected and empty tqueues */
    for (tq = empty_tqueues; tq; tq = next_tq) {
	next_tq = tq->next;
	
	if (!tq->connected)
	    free_tqueue(tq);
    }
}

extern Var call_verb8();

void
run_server_task(Objid player, Objid what, const char *verb, Var args,
		Var *result)
{
    Verbdef    *verbdef;
    Objid vloc;

    verbdef = find_verb(what, verb, &vloc);
    if (verbdef && verbdef->program) {
	Var	r;

	current_task_id = new_task_id();
	r = do_server_task(what, verb, args, vloc, verbdef->name,
		       verbdef->program, verbdef->owner,
		       verbdef->perms & (1<<VP_DEBUG), player, -1);
	if (result)
	    *result = r;
	else
	    free_var(r);
    }
}

int
task_id()
{
    return current_task_id;
}

static Var
list_for_forked_task(forked_task ft) 
{
    Var		list;

    list = new_list(9);
    list.v.list[1].type = TYPE_NUM;
    list.v.list[1].v.num = ft.id;
    list.v.list[2].type = TYPE_NUM;
    list.v.list[2].v.num = ft.start_time;
    list.v.list[3].type = TYPE_NUM;
    list.v.list[3].v.num = 0;		/* OBSOLETE: was clock ID */
    list.v.list[4].type = TYPE_NUM;
    list.v.list[4].v.num = BG_TICKS;	/* OBSOLETE: was clock ticks */
    list.v.list[5].type = TYPE_OBJ;
    list.v.list[5].v.obj = ft.a.progr;
    list.v.list[6].type = TYPE_OBJ;
    list.v.list[6].v.obj = ft.a.vloc;
    list.v.list[7].type = TYPE_STR;
    list.v.list[7].v.str = strdup(ft.a.verbname);
    list.v.list[8].type = TYPE_NUM;
    list.v.list[8].v.num = get_lineno(ft.program, 0, ft.f_index);
    list.v.list[9].type = TYPE_OBJ;	
    list.v.list[9].v.obj = ft.a.this;

    return list;
}

static Var
list_for_vm(Var partial_list, vm the_vm)
{ /* code factoring: list_for_suspended_task and list_for_reading_task */
    partial_list.v.list[5].type = TYPE_OBJ;
    partial_list.v.list[5].v.obj = progr_of_cur_verb(the_vm);
    partial_list.v.list[6].type = TYPE_OBJ;
    partial_list.v.list[6].v.obj = top_activ(the_vm).vloc;
    partial_list.v.list[7].type = TYPE_STR;
    partial_list.v.list[7].v.str = strdup(top_activ(the_vm).verbname);
    partial_list.v.list[8].type = TYPE_NUM;
    partial_list.v.list[8].v.num = suspended_lineno_of_vm(the_vm);
    partial_list.v.list[9].type = TYPE_OBJ;	
    partial_list.v.list[9].v.obj = top_activ(the_vm).this;

    return partial_list;
}

static Var
list_for_suspended_task(suspended_task st)
{
    Var		list;

    list = new_list(9);
    list.v.list[1].type = TYPE_NUM;
    list.v.list[1].v.num = st.id;
    list.v.list[2].type = TYPE_NUM;
    list.v.list[2].v.num = st.start_time;
    list.v.list[3].type = TYPE_NUM;
    list.v.list[3].v.num = 0;		/* OBSOLETE: was clock ID */
    list.v.list[4].type = TYPE_NUM;
    list.v.list[4].v.num = BG_TICKS;	/* OBSOLETE: was clock ticks */

    return list_for_vm(list, st.the_vm);
}

static Var
list_for_reading_task(Objid player, int reading_task_id, vm the_vm)
{
    Var		list;

    list = new_list(9);
    list.v.list[1].type = TYPE_NUM;
    list.v.list[1].v.num = reading_task_id;
    list.v.list[2].type = TYPE_NUM;
    list.v.list[2].v.num = -1;		/* conventional value */
    list.v.list[3].type = TYPE_NUM;
    list.v.list[3].v.num = 0;		/* OBSOLETE: was clock ID */
    list.v.list[4].type = TYPE_NUM;
    list.v.list[4].v.num = BG_TICKS;	/* OBSOLETE: was clock ticks */

    list = list_for_vm(list, the_vm);

    list.v.list[5].v.obj = player;

    return list;
}

Var
queued_tasks(Objid owner)
{	/* works for FORKED, READING and SUSPENDED tasks */   
    Var		tasks;
    int		show_all = is_wizard(owner);
    tqueue     *tq;
    task       *t;
    int		i, count = 0;

    for (tq = active_tqueues; tq; tq = tq->next) {
	if (tq->reading && (show_all  ||  tq->player == owner))
	    count++;

	for (t = tq->first; t; t = t->next)
	    if (t->kind == TASK_FORKED  &&  (show_all  ||
					     t->t.forked.a.progr == owner))
		count++;
	    else if (t->kind == TASK_SUSPENDED
		     && (progr_of_cur_verb(t->t.suspended.the_vm) == owner
			 ||  show_all))
		count++;
    }		    

    for (t = waiting_tasks; t; t = t->next) 
	if (t->kind == TASK_FORKED  &&  (show_all  ||
					 t->t.forked.a.progr == owner))
	    count++;
	else if (t->kind == TASK_SUSPENDED
		 && (progr_of_cur_verb(t->t.suspended.the_vm) == owner
		     ||  show_all))
	    count++;
    
    tasks = new_list(count);
    i = 1;

    for (tq = active_tqueues; tq; tq = tq->next) {

	if (tq->reading && (show_all  ||  tq->player == owner))
	    tasks.v.list[i++] = list_for_reading_task(tq->player,
						      tq->reading_task_id,
						      tq->reading_vm);

	for (t = tq->first; t; t = t->next)
	    if (t->kind == TASK_FORKED  &&  (show_all  ||
					     t->t.forked.a.progr == owner))
		tasks.v.list[i++] = list_for_forked_task(t->t.forked);
	    else if (t->kind == TASK_SUSPENDED
		     && (progr_of_cur_verb(t->t.suspended.the_vm) == owner
			 ||  show_all))
		tasks.v.list[i++] = list_for_suspended_task(t->t.suspended);
    }		    

    for (t = waiting_tasks; t; t = t->next) {
	if (t->kind == TASK_FORKED  &&  (show_all  ||
					 t->t.forked.a.progr == owner))
	    tasks.v.list[i++] = list_for_forked_task(t->t.forked);
	else if (t->kind == TASK_SUSPENDED
		 && (progr_of_cur_verb(t->t.suspended.the_vm) == owner
		     ||  show_all))
	    tasks.v.list[i++] = list_for_suspended_task(t->t.suspended);
    }

    return tasks;
}

enum error
kill_task(int id, Objid owner)
{ 	/* works for FORKED, READING and SUSPENDED tasks */
    task      **tt;
    tqueue     *tq;

    if (id == current_task_id) {
	abort_running_task();
	return E_NONE;
    }

    for (tt = &waiting_tasks; *tt; tt = &((*tt)->next)) {
	task   *t = *tt;
	
	if (t->kind == TASK_FORKED && t->t.forked.id == id) {
	    if (!is_wizard(owner)  &&  owner != t->t.forked.a.progr)
		return E_PERM;
	    *tt = t->next;
	    free_task(t, 1);
	    return E_NONE;
	} else if (t->kind == TASK_SUSPENDED
		   &&  t->t.suspended.id == id) {
	    if (!is_wizard(owner)  &&  
		owner != progr_of_cur_verb(t->t.suspended.the_vm))
		return E_PERM;
	    *tt = t->next;
	    free_task(t, 1);
	    return E_NONE;
	}
    }

    for (tq = active_tqueues; tq; tq = tq->next) {

	if (tq->reading && tq->reading_task_id == id) {
	    if (!is_wizard(owner) && owner != tq->player)
		return E_PERM;
	    free_vm(tq->reading_vm, 1);
	    tq->reading = 0;
	    return E_NONE;
	}	
	
	for (tt = &(tq->first); *tt; tt = &((*tt)->next)) {
	    task   *t = *tt;
	    
	    if (t->kind == TASK_FORKED  &&  t->t.forked.id == id) {
		if (!is_wizard(owner)  &&  owner != tq->player)
		    return E_PERM;
		*tt = t->next;
		if (t->next == 0)
		    tq->last = tt;
		free_task(t, 1);
		return E_NONE;
	    } else if (t->kind == TASK_SUSPENDED && t->t.suspended.id == id) {
		if (!is_wizard(owner)  &&  
		    owner != progr_of_cur_verb(t->t.suspended.the_vm))
		    return E_PERM;
		*tt = t->next;
		if (t->next == 0) 
		    tq->last = tt;
		free_task(t, 1);
		return E_NONE;
	    }
	}
    }

    return E_INVARG;
}

int
output_delimiters(Objid player, const char **prefix, const char **suffix)
{
    tqueue     *tq = find_tqueue(player, 0);

    if (!tq || !tq->connected)
	return 0;

    if (tq->output_prefix)
	*prefix = tq->output_prefix;
    else
	*prefix = "";

    if (tq->output_suffix)
	*suffix = tq->output_suffix;
    else
	*suffix = "";

    return 1;
}

static void
write_forked_task(forked_task ft)
{
    /* losing some info because not all of the program written */
    int lineno = get_lineno(ft.program, 0, ft.f_index);

    db_printf("0 %d %d %d\n", lineno, ft.start_time, ft.id);
    write_activ_as_pi(ft.a);
    write_rt_env(ft.program->var_names, ft.rt_env, ft.program->num_var_names);
    write_forked_program(ft.program, ft.f_index);
}

static void
write_suspended_task(suspended_task st)
{
    db_printf("%d %d\n", st.start_time, st.id);
    write_vm(st.the_vm);
}

void
write_task_queue(FILE *f)
{
    int		forked_count = 0;
    int		suspended_count = 0;
    task       *t;
    tqueue     *tq;

    db_printf("0 clocks\n");	/* for compatibility's sake */

    for (t = waiting_tasks; t; t = t->next)
	if (t->kind == TASK_FORKED)
	    forked_count++;
	else /* t->kind == TASK_SUSPENDED */
	    suspended_count++;
    
    for (tq = active_tqueues; tq; tq = tq->next)
	for (t = tq->first; t; t = t->next)
	    if (t->kind == TASK_FORKED)
		forked_count++;
	    else if (t->kind == TASK_SUSPENDED)
		suspended_count++;
    

    db_printf("%d queued tasks\n", forked_count); /***** has to change name
						     to forked tasks someday */

    for (t = waiting_tasks; t; t = t->next)
	if (t->kind == TASK_FORKED)
	    write_forked_task(t->t.forked);

    for (tq = active_tqueues; tq; tq = tq->next)
	for (t = tq->first; t; t = t->next)
	    if (t->kind == TASK_FORKED)
		write_forked_task(t->t.forked);

    db_printf("%d suspended tasks\n", suspended_count);
    
    for (t = waiting_tasks; t; t = t->next)
	if (t->kind == TASK_SUSPENDED)
	    write_suspended_task(t->t.suspended);

    for (tq = active_tqueues; tq; tq = tq->next)
	for (t = tq->first; t; t = t->next)
	    if (t->kind == TASK_SUSPENDED)
		write_suspended_task(t->t.suspended);

}

int
read_task_queue(FILE *f)
{
    int		count, suspended_count, suspended_task_header;

    /* Skip obsolete clock stuff */
    if (fscanf(f, "%d clocks\n", &count) != 1) {
        errlog("READ_TASK_QUEUE: Bad clock count.\n");
        return -1;
    }
    for (; count > 0; count--) {
	int	dummy;

	if (fscanf(f, "%d %d %d\n", &dummy, &dummy, &dummy) != 3) {
            errlog("READ_TASK_QUEUE: Bad clock; count = %d\n", count);
            return -1;
        }
    }

    if (fscanf(f, "%d queued tasks\n", &count) != 1) {
        errlog("READ_TASK_QUEUE: Bad task count.\n");
        return -1;
    }
    for (; count > 0; count--) {
	int		i, first_lineno, id, old_size;
	time_t		start_time;
	Program	       *program;
	Var	       *rt_env, *old_rt_env;
	char	      **old_names;
	activation      a;

	if (fscanf(f, "%d %d %d %d", &i, &first_lineno, &start_time, &id) != 4
	    || fgetc(f) != '\n') {
	    errlog("READ_TASK_QUEUE: Bad numbers, count = %d.\n", count);
            return -1;
        }
	if (read_activ_as_pi(f, &a) < 0) {
            errlog("READ_TASK_QUEUE: Bad activation, count = %d.\n", count);
            return -1;
        }
        if (read_rt_env(f, &old_names, &old_rt_env, &old_size) < 0) {
            errlog("READ_TASK_QUEUE: Bad env, count = %d.\n", count);
            return -1;
        }
        if (read_program(f, &program) < 0) {
            errlog("READ_TASK_QUEUE: Bad program, count = %d.\n", count);
            return -1;
        }
	rt_env = reorder_rt_env(old_rt_env, old_names, old_size, program);
	program->first_lineno = first_lineno;
	
	enqueue_ft(program, a, rt_env, MAIN_VECTOR, start_time, id);
	/***** wierd! the forked statement is parsed as the main vector
	  hence this is a forked task but running on the MAIN vector */
    }	

    suspended_task_header = fscanf(f, "%d suspended tasks\n", &suspended_count);
    if (suspended_task_header == EOF)
	return 0; /* old version */

    if (suspended_task_header != 1) {
	errlog("READ_TASK_QUEUE: Bad suspended task count.\n");
	return -1;
    }

    for (; suspended_count > 0; suspended_count--) {
	task       *t = (task *) mymalloc(sizeof(task), M_TASK);

	t->kind = TASK_SUSPENDED;
	if (fscanf(f, "%d %d\n", &(t->t.suspended.start_time), 
		                 &(t->t.suspended.id)) != 2) {
	    errlog("READ_TASK_QUEUE: Bad suspended task header. count = %d\n",
		   suspended_count);
	    return -1;
	}

	if (read_vm(f, &(t->t.suspended.the_vm)) < 0) {
	    errlog("READ_TASK_QUEUE: Bad suspended task vm. count = %d\n",
		   suspended_count);
	    return -1;
	}

	enqueue_waiting(t);
    }
    return 0;
}

/* $Log: tasks.c,v $
 * Revision 1.5  1992/08/12  01:51:16  pjames
 * Copied verb and verbname when creating a forked_task.
 *
 * Revision 1.4  1992/08/10  16:50:03  pjames
 * Updated #includes.  Changed forked_task data structure to hold an
 * activation instead of a Parse_Info, and all manipulation of
 * forked_task.pi now uses correct fields from forked_task.a;
 *
 * Revision 1.3  1992/07/27  18:24:33  pjames
 * Change ct_env_size to num_var_names.
 *
 * Revision 1.2  1992/07/21  00:07:17  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 * */
