char rcsid_unparse[] = "$Id: unparse.c,v 1.7 1992/08/14 00:26:51 pavel Exp $";

#include <ctype.h>
#include "ast.h"
#include "exceptions.h"
#include "log.h"
#include "opcode.h"
#include "unparse.h"
#include "storage.h"
#include "utils.h"

static Program *prog;

/* pseudo-properties; all objects have these */

static struct pptable_entry  pptable[] = {
    "programmer",	FLAG - FLAG_PROGRAMMER,	TYPE_NUM, 1,
    "wizard",		FLAG - FLAG_WIZARD,	TYPE_NUM, 1,
    "r",		FLAG - FLAG_READ,	TYPE_NUM, 2,
    "w",		FLAG - FLAG_WRITE,	TYPE_NUM, 2,
    "location",		LOCATION,		TYPE_NUM, 0,
    "owner",		OWNER,			TYPE_NUM, 0,
    "name",		NAME,			TYPE_STR, 2,
    "contents",		CONTENTS,		TYPE_LIST, 0,
};

struct pptable_entry  *
match_pseudo_property(const char *pname, Propid *pid)
{
    int		i;

    for (i = 0; i < Arraysize(pptable, struct pptable_entry); i++) {
        if (!mystrcasecmp(pname, pptable[i].name)) {
            *pid = pptable[i].propno;
            return &pptable[i];
        }
    }
    return 0;
}

struct err_idx {
    enum error	err;
    const char	*name;
};

static struct err_idx errors[] = {
    E_NONE,	"No error",
    E_TYPE,	"Type mismatch",
    E_DIV,	"Division by zero",
    E_PERM,	"Permission denied",
    E_PROPNF,	"Property not found",
    E_VERBNF,	"Verb not found",
    E_VARNF,	"Variable not found",
    E_INVIND,	"Invalid indirection",
    E_RECMOVE,	"Recursive move",
    E_MAXREC,	"Too many verb calls",
    E_RANGE,	"Range error",
    E_ARGS,	"Incorrect number of arguments",
    E_NACC,	"Move refused by destination",
    E_INVARG,	"Invalid argument",
    E_QUOTA,	"Object ownership quota exceeded",
};

const char *
unparse_error(enum error errno)
{
    int		i;

    for (i = 0; i < Arraysize(errors, struct err_idx); i++) {
	if (errno == errors[i].err)
	    return errors[i].name;
    }
    return "Unknown Error";
}

const char *
unparse_option(int option)
{
    switch(option) {
      case OPT_NONE:
	return "none";
      case OPT_ANY:
	return "any";
      case OPT_THIS:
	return "this";
      default:
	return "unparse error";
    }
}

const char *
unparse_prep(int prep)
{
    if (prep == NOTHING)
	return "none";
    else if (prep == AMBIGUOUS)
	return "any";
    else
	return prep_name(prep);
}

struct prec {
    enum Expr_Kind	kind;
    int			precedence;
};

static struct prec prec_table[] = {
    EXPR_ASGN,	1,
    EXPR_COND,	2,	/* the unparser for this depends on only ASGN having
			   lower precedence.  Fix that if this changes. */
    EXPR_OR,	3,
    EXPR_AND,	3,
    EXPR_EQ,	4,
    EXPR_NE,	4,
    EXPR_LT,	4,
    EXPR_LE,	4,
    EXPR_GT,	4,
    EXPR_GE,	4,
    EXPR_IN,	4,
    EXPR_PLUS,	5,
    EXPR_MINUS,	5,
    EXPR_TIMES,	6,
    EXPR_DIVIDE,6,
    EXPR_MOD,	6,
    EXPR_NEGATE,7,
    EXPR_NOT,	7,
    EXPR_PROP,	8,
    EXPR_VERB,	8,
    EXPR_INDEX,	8,
    EXPR_RANGE,	8,
    EXPR_VAR,	9,
    EXPR_ID,	9,
    EXPR_LIST,	9,
    EXPR_CALL,	9
};

static int expr_prec[SizeOf_Expr_Kind];

struct binop {
    enum Expr_Kind	kind;
    const char	       *string;
};

static struct binop binop_table[] = {
    EXPR_IN,	" in ",
    EXPR_OR,	" || ",
    EXPR_AND,	" && ",
    EXPR_EQ,	" == ",
    EXPR_NE,	" != ",
    EXPR_LT,	" < ",
    EXPR_LE,	" <= ",
    EXPR_GT,	" > ",
    EXPR_GE,	" >= ",
    EXPR_PLUS,	" + ",
    EXPR_MINUS,	" - ",
    EXPR_TIMES,	" * ",
    EXPR_DIVIDE," / ",
    EXPR_MOD,	" % "
};

static const char *binop_string[SizeOf_Expr_Kind];

static int expr_tables_initialized = 0;

static void
init_expr_tables()
{
    int i;

    for (i = 0; i < Arraysize(prec_table, struct prec); i++)	
	expr_prec[prec_table[i].kind] = prec_table[i].precedence;

    for (i = 0; i < Arraysize(binop_table, struct binop); i++)
	binop_string[binop_table[i].kind] = binop_table[i].string;

    expr_tables_initialized = 1;
}

/********** globals *********************************/

static Unparser_Receiver	receiver;
static void		       *receiver_data;
static int			fully_parenthesize, indent_code;

/********** AST to receiver procedures **************/

static void unparse_stmt(Stmt *s, int indent);
static void unparse_expr(Stream *str, Expr *e);
static void unparse_arglist(Stream *str, Arg_List *a);
static void unparse_asgn(Stream *str, struct Expr_Asgn asgn);

void
list_prg(Stmt *program, int p, int i)
{
    fully_parenthesize = p;
    indent_code = i;
    if (! expr_tables_initialized)
	init_expr_tables();
    unparse_stmt(program, 0);
}

static void
bracket_lt(Stream *str, enum Expr_Kind parent, Expr *child)
{
    if ((fully_parenthesize && expr_prec[child->kind] < expr_prec[EXPR_PROP])
	||  expr_prec[parent] > expr_prec[child->kind]) {
	stream_add_char(str, '(');
	unparse_expr(str, child);
	stream_add_char(str, ')');
    } else {
	unparse_expr(str, child);
    }
}

static void
bracket_le(Stream *str, enum Expr_Kind parent, Expr *child)
{
    if ((fully_parenthesize && expr_prec[child->kind] < expr_prec[EXPR_PROP])
	||  expr_prec[parent] >= expr_prec[child->kind]) {
	stream_add_char(str, '(');
	unparse_expr(str, child);
	stream_add_char(str, ')');
    } else {
	unparse_expr(str, child);
    }
}

static void
output(Stream *str)
{
    (*receiver)(receiver_data, reset_stream(str));
}

static void
indent_stmt(Stream *str, int indent)
{
    int		i;

    if (indent_code)
	for (i = 0; i < indent; i++)
	    stream_add_char(str, ' ');
}

static void
unparse_stmt_cond(Stream *str, struct Stmt_Cond cond, int indent)
{
    Cond_Arm       *elseifs;
	
    stream_add_string(str, "if (");
    unparse_expr(str, cond.arms->condition);
    stream_add_char(str, ')');
    output(str);
    unparse_stmt(cond.arms->stmt, indent + 2);
    for (elseifs = cond.arms->next; elseifs; elseifs = elseifs->next) {
	indent_stmt(str, indent);
	stream_add_string(str, "elseif (");
	unparse_expr(str, elseifs->condition);
	stream_add_char(str, ')');
	output(str);
	unparse_stmt(elseifs->stmt, indent + 2);
    }
    if (cond.otherwise) {
	indent_stmt(str, indent);
	stream_add_string(str, "else");
	output(str);
	unparse_stmt(cond.otherwise, indent + 2);
    }
    indent_stmt(str, indent);
    stream_add_string(str, "endif");
    output(str);
}

static void
unparse_stmt_list(Stream *str, struct Stmt_List list, int indent)
{
    stream_printf(str, "for %s in (", prog->var_names[list.id.slot]);
    unparse_expr(str, list.expr);
    stream_add_char(str, ')');
    output(str);
    unparse_stmt(list.body, indent + 2);
    indent_stmt(str, indent);
    stream_add_string(str, "endfor");
    output(str);
}

static void
unparse_stmt_range(Stream *str, struct Stmt_Range range, int indent)
{
    stream_printf(str, "for %s in [", prog->var_names[range.id.slot]);
    unparse_expr(str, range.from);
    stream_add_string(str, "..");
    unparse_expr(str, range.to);
    stream_add_char(str, ']');
    output(str);
    unparse_stmt(range.body, indent + 2);
    indent_stmt(str, indent);
    stream_add_string(str, "endfor");
    output(str);
}

static void
unparse_stmt_fork(Stream *str, struct Stmt_Fork fork_stmt, int indent)
{
    if (fork_stmt.id.slot != -1)
	stream_printf(str, "fork %s (", prog->var_names[fork_stmt.id.slot]);
    else
	stream_add_string(str, "fork (");
    unparse_expr(str, fork_stmt.time);
    stream_add_char(str, ')');
    output(str);
    unparse_stmt(fork_stmt.body, indent + 2);
    indent_stmt(str, indent);
    stream_add_string(str, "endfork");
    output(str);
}

static void
unparse_stmt(Stmt *stmt, int indent)
{
    Stream     *str = new_stream(100);

    while (stmt) {
	indent_stmt(str, indent);
	switch(stmt->kind) {
	  case STMT_COND:
	    unparse_stmt_cond(str, stmt->s.cond, indent);
	    break;
	  case STMT_LIST:
	    unparse_stmt_list(str, stmt->s.list, indent);
	    break;
	  case STMT_RANGE:
	    unparse_stmt_range(str, stmt->s.range, indent);
	    break;
	  case STMT_FORK:
	    unparse_stmt_fork(str, stmt->s.fork, indent);
	    break;
	  case STMT_EXPR:
	    unparse_expr(str, stmt->s.expr);
	    stream_add_char(str, ';');
	    output(str);
	    break;
	  case STMT_WHILE:
	    stream_add_string(str, "while (");
	    unparse_expr(str, stmt->s.loop.condition);
	    stream_add_char(str, ')');
	    output(str);
	    unparse_stmt(stmt->s.loop.body, indent + 2);
	    indent_stmt(str, indent);
	    stream_add_string(str, "endwhile");
	    output(str);
	    break;
	  case STMT_RETURN:
	    if (stmt->s.expr) {
		stream_add_string(str, "return ");
		unparse_expr(str, stmt->s.expr);
	    } else
		stream_add_string(str, "return");
	    stream_add_char(str, ';');
	    output(str);
	    break;
	  default:
	    errlog("UNPARSE_STMT: Unknown Stmt_Kind: %d\n", stmt->kind);
	    stream_add_string(str, "?!?!?!?;");
	    output(str);
	    break;
	}
	stmt = stmt->next;
    }

    free_stream(str);
}

static int
ok_identifier(const char *name)
{
    if (*name != '\0'  &&  (isalpha(*name) || *name == '_')) {
	while (*++name != '\0'  &&  (isalnum(*name) || *name == '_'))
	    ;
	if (*name == '\0')
	    return 1;
    }

    return 0;
}

static void
unparse_name_expr(Stream *str, Expr *expr)
{
    /*
     * Handle the right-hand expression in EXPR_PROP and EXPR_VERB.
     * If it's a simple string literal with the syntax of an identifier,
     * just print the name.  Otherwise, use parens and unparse the
     * expression normally.
     */

    if (expr->kind == EXPR_VAR  &&  expr->e.var.type == TYPE_STR
	                        &&  ok_identifier(expr->e.var.v.str)) {
	stream_add_string(str, expr->e.var.v.str);
	return;
    }

    /* We need to use the full unparser */
    stream_add_char(str, '(');
    unparse_expr(str, expr);
    stream_add_char(str, ')');
}

static void
unparse_expr(Stream *str, Expr *expr)
{
    switch(expr->kind) {
      case EXPR_PROP:
	if (expr->e.bin.lhs->kind == EXPR_VAR
	    && expr->e.bin.lhs->e.var.type == TYPE_OBJ
	    && expr->e.bin.lhs->e.var.v.obj == 0
	    && expr->e.bin.rhs->kind == EXPR_VAR
	    && expr->e.bin.rhs->e.var.type == TYPE_STR
	    && ok_identifier(expr->e.bin.rhs->e.var.v.str)) {
	    stream_add_char(str, '$');
	    stream_add_string(str, expr->e.bin.rhs->e.var.v.str);
	} else {
	    bracket_lt(str, EXPR_PROP, expr->e.bin.lhs);
	    stream_add_char(str, '.');
	    unparse_name_expr(str, expr->e.bin.rhs);
	}
	break;

      case EXPR_VERB:
	bracket_lt(str, EXPR_VERB, expr->e.verb.obj);
	stream_add_char(str, ':');
	unparse_name_expr(str, expr->e.verb.verb);
	stream_add_char(str, '(');
	unparse_arglist(str, expr->e.verb.args);
	stream_add_char(str, ')');
	break;

      case EXPR_INDEX:
	bracket_lt(str, EXPR_INDEX, expr->e.bin.lhs);
	stream_add_char(str, '[');
	unparse_expr(str, expr->e.bin.rhs);
	stream_add_char(str, ']');
	break;

      case EXPR_RANGE:
	bracket_lt(str, EXPR_RANGE, expr->e.range.base);
	stream_add_char(str, '[');
	unparse_expr(str, expr->e.range.from);
	stream_add_string(str, "..");
	unparse_expr(str, expr->e.range.to);
	stream_add_char(str, ']');
	break;

      case EXPR_PLUS:	case EXPR_MINUS:	case EXPR_TIMES:
      case EXPR_DIVIDE:	case EXPR_MOD:		case EXPR_AND:
      case EXPR_OR:	case EXPR_EQ:		case EXPR_NE:
      case EXPR_LT:	case EXPR_GT:		case EXPR_LE:
      case EXPR_GE:	case EXPR_IN:
	bracket_lt(str, expr->kind, expr->e.bin.lhs);
	stream_add_string(str, binop_string[expr->kind]);
	bracket_le(str, expr->kind, expr->e.bin.rhs);
	break;

      case EXPR_COND:
	bracket_le(str, EXPR_COND, expr->e.cond.condition);
	stream_add_string(str, " ? ");
	unparse_expr(str, expr->e.cond.consequent);
	stream_add_string(str, " | ");
	bracket_le(str, EXPR_COND, expr->e.cond.alternate);
	break;
	
      case EXPR_NEGATE:
	stream_add_char(str, '-');
	bracket_lt(str, EXPR_NEGATE, expr->e.expr);
	break;
	
      case EXPR_NOT:
	stream_add_char(str, '!');
	bracket_lt(str, EXPR_NOT, expr->e.expr);
	break;

      case EXPR_VAR:
	switch (expr->e.var.type) {
	  case TYPE_NUM:
	    stream_printf(str, "%d", expr->e.var.v.num);
	    break;
	  case TYPE_OBJ:
	    stream_printf(str, "#%d", expr->e.var.v.obj);
	    break;
	  case TYPE_STR:
	    {
		char *s = expr->e.var.v.str;
		
		stream_add_char(str, '"');
		while (*s) {
		    switch (*s) {
		      case '"':
		      case '\\':
			stream_add_char(str, '\\');
			/* fall thru */
		      default:
			stream_add_char(str, *s++);
		    }
		}
		stream_add_char(str, '"');
	    }
	    break;
	  case TYPE_ERR:
	    {
		struct entry {
		    const char *name;
		    enum error	err;
		};
		static struct entry	entries[] = {
		    { "E_NONE", E_NONE },
		    { "E_TYPE", E_TYPE },
		    { "E_DIV", E_DIV },
		    { "E_PERM", E_PERM },
		    { "E_PROPNF", E_PROPNF },
		    { "E_VERBNF", E_VERBNF },
		    { "E_VARNF", E_VARNF },
		    { "E_INVIND", E_INVIND },
		    { "E_RECMOVE", E_RECMOVE },
		    { "E_MAXREC", E_MAXREC },
		    { "E_RANGE", E_RANGE },
		    { "E_ARGS", E_ARGS },
		    { "E_NACC", E_NACC },
		    { "E_INVARG", E_INVARG },
		    { "E_QUOTA", E_QUOTA },
		};
		int		i;
		enum error	err = expr->e.var.v.err;

		for (i = 0; i < Arraysize(entries, struct entry); i++) {
		    if (entries[i].err == err) {
			stream_add_string(str, entries[i].name);
			break;
		    }
		}
		if (i >= Arraysize(entries, struct entry)) {
		    errlog("UNPARSE_EXPR: Unknown enum error: %d\n", err);
		    stream_add_string(str, "E_?!?!?");
		}
	    }
	    break;
	  default:
	    errlog("UNPARSE_EXPR: Unknown EXPR_VAR type: %d\n",
		   expr->e.var.type);
	    stream_add_string(str, ">>Unknown-literal<<");
	    break;
	}
	break;

      case EXPR_ASGN:
	unparse_asgn(str, expr->e.asgn);
	break;

      case EXPR_CALL:
	stream_add_string(str, name_func_by_num(expr->e.call.func));
	stream_add_char(str, '(');
	unparse_arglist(str, expr->e.call.args);
	stream_add_char(str, ')');
	break;

      case EXPR_ID:
	stream_add_string(str, prog->var_names[expr->e.id.slot]);
	break;

      case EXPR_LIST:
	stream_add_char(str, '{');
	unparse_arglist(str, expr->e.list);
	stream_add_char(str, '}');
	break;

      default:
	errlog("UNPARSE_EXPR: Unknown Expr_Kind: %d\n", expr->kind);
	stream_add_string(str, "(?!?!?!?!?)");
	break;
    }
}

static void
unparse_asgn(Stream *str, struct Expr_Asgn asgn)
{
    switch (asgn.kind) {
      case ASGN_VAR:
	stream_add_string(str, prog->var_names[asgn.a.id.slot]);
	break;

      case ASGN_PROP:
	if (asgn.a.bin.lhs->kind == EXPR_VAR
	    && asgn.a.bin.lhs->e.var.type == TYPE_OBJ
	    && asgn.a.bin.lhs->e.var.v.obj == 0
	    && asgn.a.bin.rhs->kind == EXPR_VAR
	    && asgn.a.bin.rhs->e.var.type == TYPE_STR
	    && ok_identifier(asgn.a.bin.rhs->e.var.v.str)) {
	    stream_add_char(str, '$');
	    stream_add_string(str, asgn.a.bin.rhs->e.var.v.str);
	} else {
	    bracket_lt(str, EXPR_PROP, asgn.a.bin.lhs);
	    stream_add_char(str, '.');
	    unparse_name_expr(str, asgn.a.bin.rhs);
	}
	break;

      case ASGN_REF:
	bracket_lt(str, EXPR_INDEX, asgn.a.bin.lhs);
	stream_add_char(str, '[');
	unparse_expr(str, asgn.a.bin.rhs);
	stream_add_char(str, ']');
	break;

      default:
	errlog("UNPARSE_ASGN: Unknown Asgn_Kind: %d\n", asgn.kind);
	stream_add_string(str, "?!?!?!?!?");
	break;
    }

    stream_add_string(str, " = ");
    bracket_lt(str, EXPR_ASGN, asgn.value);
}

static void
unparse_arglist(Stream *str, Arg_List *args)
{
    while (args) {
	if (args->kind == ARG_SPLICE)
	    stream_add_char(str, '@');
	unparse_expr(str, args->expr);
	if (args->next)
	    stream_add_string(str, ", ");
	args = args->next;
    }
}

enum lineno_position {TOP, BOTTOM};

/*** b2ast globals ***/

static void *hot_statement;	/* the AST branch the pc is on */
static enum lineno_position lp; /* where the lineno is in the branch */
static int is_hot_statement;	/* boolean: is this the desired statement? */
static unsigned hot_byte;	/* desired byte for lineno */
static unsigned current_vector;	/* keeps track of which vector we are in */
static unsigned desired_vector;	/* keeps track of which vector we want */

static Expr **expr_stack;
static unsigned top_expr_stack;
static Stmt **else_stack;
static unsigned top_else_stack;
static unsigned max_else_stack;
static Cond_Arm **cond_arm_stack;
static unsigned top_cond_arm_stack;
static unsigned max_cond_arm_stack;

static int expr2ast(Bytecodes, unsigned, unsigned);
static int stmt2ast(Bytecodes, unsigned, unsigned);

#define POP_EXPR() expr_stack[--top_expr_stack]
#define PUSH_EXPR(e) expr_stack[top_expr_stack++] = e

#define CURR_ELSE else_stack[top_else_stack - 1]
#define POP_ELSE() else_stack[--top_else_stack]

void
PUSH_ELSE(Stmt *s)
{
    if (top_else_stack == max_else_stack) {
	int loop;
	unsigned new_size = (top_else_stack ? 2 * top_else_stack : 5);
        Stmt **nes = mymalloc(sizeof(Stmt *) * new_size, M_STMT_STACK);

	if (else_stack != NULL) {
	    for (loop = 0; loop < top_else_stack; loop++)
		nes[loop] = else_stack[loop];
	    myfree(else_stack, M_STMT_STACK);
	}
	max_else_stack = new_size;
        else_stack = nes;
    }
    else_stack[top_else_stack++] = s;
}
    
#define CURR_COND_ARM cond_arm_stack[top_cond_arm_stack - 1]

static void
PUSH_STMT(Stmt *stmt, int hot, enum lineno_position where)
{
    Stmt *index = CURR_COND_ARM->stmt;
    if (index == NULL)
	CURR_COND_ARM->stmt = stmt;
    else {
	while (index->next != NULL)
	    index = index->next;
	index->next = stmt;
    }
    stmt->next = NULL;
    if (hot_statement == NULL && hot) {
	hot_statement = (void *)stmt;
	lp = where;
    }
}

#define POP_COND_ARM() cond_arm_stack[--top_cond_arm_stack]

static void
PUSH_COND_ARM(Cond_Arm *ca)
{
    if (top_cond_arm_stack == max_cond_arm_stack) {
	int loop;
	unsigned new_size = (top_cond_arm_stack ? 2 * top_cond_arm_stack : 5);
	Cond_Arm ** new_ca_stack =
	    mymalloc(sizeof(Cond_Arm *) * new_size, M_COND_ARM_STACK);

	if (cond_arm_stack != NULL) {
	    for (loop = 0; loop < top_cond_arm_stack; loop++)
		new_ca_stack[loop] = cond_arm_stack[loop];
   	    myfree(cond_arm_stack, M_COND_ARM_STACK);
	}
	max_cond_arm_stack = new_size;
	cond_arm_stack = new_ca_stack;
    }
    cond_arm_stack[top_cond_arm_stack++] = ca;
}

/*** failure handler ***/

static void
unparse_fail()
{
    panic("Unparse failed!");
}

/* Byte *bv = bytecodes.vector + PREVIOUS address, */
/* unsigned nb = number of bytes to read           */
#define READ_BYTES_FROM_VECTOR(bv, nb)	\
  ( (nb == 1)				\
    ? *(bv+1)				\
    : ( (nb == 2)			\
        ? ((unsigned) *(bv+1) << 8)	\
          + *(bv+2)			\
        : ((unsigned) *(bv+1) << 24)	\
          + ((unsigned) *(bv+2) << 16)	\
          + ((unsigned) *(bv+3) << 8)	\
          + *(bv+4)))

Stmt
*b2ast(Bytecodes b, unsigned start, unsigned end)
{
    Cond_Arm *ca;
    Stmt *ans;
    
    else_stack = NULL;
    top_else_stack = 0;
    max_else_stack = 0;
    
    cond_arm_stack = NULL;
    top_cond_arm_stack = 0;
    max_cond_arm_stack = 0;
    
    hot_statement = NULL;
    is_hot_statement = 0;

    current_vector = MAIN_VECTOR;
    top_expr_stack = 0;
    (void)stmt2ast(b, start, end);
    
    ca = POP_COND_ARM();
    ans = ca->stmt;
    myfree(ca, M_AST);

    if (else_stack != NULL)
	myfree(else_stack, M_STMT_STACK);
    if (cond_arm_stack != NULL)
	myfree(cond_arm_stack, M_COND_ARM_STACK);
    
    return ans;
}

static int
stmt2ast(Bytecodes b, unsigned start, unsigned end)
{
    /* parse stmt's and put them on the current cond arm until end   */
    /* is reached.  False cond arm placed on stack for 'main', fork, */
    /* and loop statements.  Returns index of first unused byte.     */

    register Byte *bv, *bv_end;
    unsigned label, label2;
    Stmt *s1;
    Expr *e1;
    Cond_Arm *ca1, *ca2;
    int was_hot;
    int hot_top;
    unsigned old_vector;
    
    /* place a phony condition arm on stack for main statement list */
    if (!top_cond_arm_stack)
        PUSH_COND_ARM(alloc_cond_arm(NULL, NULL));

    for (bv = b.vector + start, bv_end = b.vector + end; bv < bv_end; bv++) {
 
       /* *** BEWARE OF bv++ DEFAULT!! ***
       most of the time, bv should be incremented by 1 for next iteration.
       Be sure to set bv to one below desired value when adjusting to a
       label */
    
        bv = b.vector + expr2ast(b, bv - b.vector, end);
        e1 = NULL;
	hot_top = 0;
	was_hot = 0;
        switch(*bv) {
	    
	  case OP_IF:
	    s1 = alloc_stmt(STMT_COND);
	    s1->s.cond.arms = alloc_cond_arm(POP_EXPR(), NULL);
	    s1->s.cond.arms->next = NULL;
	    s1->s.cond.otherwise = NULL;
	    if (current_vector == desired_vector &&
		hot_byte >= bv - b.vector &&
		hot_byte <= bv - b.vector + b.numbytes_label)
		is_hot_statement = 1;
	    label = READ_BYTES_FROM_VECTOR(bv, b.numbytes_label);
	    PUSH_COND_ARM(s1->s.cond.arms);
	    was_hot = is_hot_statement;
  	    is_hot_statement = 0;
	    if (was_hot)
		hot_top = 1;
	    bv = b.vector + stmt2ast(b, bv - b.vector + b.numbytes_label + 1,
				  label - b.numbytes_label - 1);
	    if (*bv != OP_JUMP)
		unparse_fail();
	    if (current_vector == desired_vector &&
		hot_byte >= (bv-b.vector) &&
		hot_byte <= (bv-b.vector+b.numbytes_label)) {
		was_hot = 1;
		hot_top = 0;
		is_hot_statement = 0;
	    }
	    label2 = READ_BYTES_FROM_VECTOR(bv, b.numbytes_label);
	    if (label != label2) {
		PUSH_ELSE(NULL); /* to collect else statements */
		PUSH_COND_ARM(alloc_cond_arm(NULL, NULL));
		                 /* to collect elseif statements */
		bv = b.vector +
		    stmt2ast(b, bv - b.vector + b.numbytes_label + 1, label2);
		ca2 = POP_COND_ARM();
		ca1 = POP_COND_ARM();
		if (ca2->condition) {
		    ca1->next = ca2;
		    s1->s.cond.otherwise = POP_ELSE();
		} else {
		    s1->s.cond.otherwise = ca2->stmt;
		    ca1->next = NULL;
		    myfree(ca2, M_AST);
		    if (POP_ELSE() != NULL)
			errlog("Lost else statements!");
		}
	    } else
		(void) POP_COND_ARM(); 
	    bv = b.vector + label2 - 1;
	    PUSH_STMT(s1, was_hot && hot_top, TOP);
	    if (was_hot && !hot_top) {
		hot_statement = s1->s.cond.arms;
		lp = BOTTOM;
	    }
  	    break;

	  case OP_EIF:
	    if (current_vector == desired_vector &&
		hot_byte >= bv - b.vector &&
		hot_byte <= bv - b.vector + b.numbytes_label)
		is_hot_statement = 1;
	    was_hot = is_hot_statement;
	    is_hot_statement = 0;
	    if (was_hot)
		hot_top = 1;
	    label = READ_BYTES_FROM_VECTOR(bv, b.numbytes_label);
	    CURR_COND_ARM->condition = POP_EXPR();
	    bv = b.vector + stmt2ast(b, bv - b.vector + b.numbytes_label + 1,
				  label - b.numbytes_label - 1);
	    if (*bv != OP_JUMP)
		unparse_fail();
	    if (current_vector == desired_vector &&
		hot_byte >= (bv-b.vector) &&
		hot_byte <= (bv-b.vector+b.numbytes_label)) {
		was_hot = 1;
		hot_top = 0;
		is_hot_statement = 0;
	    }
	    label2 = READ_BYTES_FROM_VECTOR(bv, b.numbytes_label);
	    if (was_hot) {
		hot_statement = (void *)cond_arm_stack[top_cond_arm_stack-1];
		lp = hot_top ? TOP : BOTTOM;
	    }
	    if (label != label2) {
		PUSH_COND_ARM(alloc_cond_arm(NULL, NULL));
		bv = b.vector +
		    stmt2ast(b, bv - b.vector + b.numbytes_label + 1, label2);
		ca2 = POP_COND_ARM();
		ca1 = POP_COND_ARM();
		if (ca2->condition) {
		    ca1->next = ca2;
		} else {
		    ca1->next = NULL;
		    CURR_ELSE = ca2->stmt;
		    myfree(ca2, M_AST);
		}
		PUSH_COND_ARM(ca1);
	    }
	    bv = b.vector + label2 - 1;
  	    break;
	    
	  case OP_WHILE:
	    s1 = alloc_stmt(STMT_WHILE);
	    if (current_vector == desired_vector &&
		hot_byte >= bv - b.vector &&
		hot_byte <= bv - b.vector + b.numbytes_label)
		is_hot_statement = 1;
	    label = READ_BYTES_FROM_VECTOR(bv, b.numbytes_label);
	    s1->s.loop.condition = POP_EXPR();
	    PUSH_COND_ARM(alloc_cond_arm(NULL, NULL)); /* dummy */
	    was_hot = is_hot_statement;
	    is_hot_statement = 0;
	    bv = b.vector + stmt2ast(b, bv-b.vector + b.numbytes_label+1,
				  label - b.numbytes_label - 1);
	    if (*bv != OP_JUMP)
		unparse_fail();
	    if (current_vector == desired_vector &&
		hot_byte >= bv - b.vector &&
		hot_byte <= bv - b.vector + b.numbytes_label)
		was_hot = 1;
	    bv += b.numbytes_label;
	    ca1 = POP_COND_ARM();
	    s1->s.loop.body = ca1->stmt;
	    myfree(ca1, M_AST);
	    PUSH_STMT(s1, was_hot, TOP);
 	    break;
	    
          case OP_FOR_LIST:
	    s1 = alloc_stmt(STMT_LIST);
	    myfree(POP_EXPR(), M_AST); /* get rid of '1' */
	    if (current_vector == desired_vector &&
		hot_byte >= bv - b.vector &&
		hot_byte <= bv - b.vector +
		b.numbytes_var_name + b.numbytes_label)
		is_hot_statement = 1;
	    s1->s.list.expr = POP_EXPR();
	    s1->s.list.id.slot =
		READ_BYTES_FROM_VECTOR(bv, b.numbytes_var_name);
	    s1->s.list.id.name = NULL;
	    bv += b.numbytes_var_name;
	    label = READ_BYTES_FROM_VECTOR(bv, b.numbytes_label);
	    PUSH_COND_ARM(alloc_cond_arm(NULL, NULL)); /* dummy */
	    was_hot = is_hot_statement;
	    is_hot_statement = 0;
	    bv = b.vector + stmt2ast(b, bv-b.vector + b.numbytes_label+1,
				  label - b.numbytes_label - 1);
	    if (current_vector == desired_vector &&
		hot_byte >= bv - b.vector &&
		hot_byte <= bv - b.vector + b.numbytes_label)
		was_hot = 1;
	    if (*bv != OP_JUMP)
		unparse_fail();
	    bv += b.numbytes_label;
	    ca1 = POP_COND_ARM();
	    s1->s.list.body = ca1->stmt;
	    myfree(ca1, M_AST);
	    PUSH_STMT(s1, was_hot, TOP);
	    break;
	    
          case OP_FOR_RANGE:
	    s1 = alloc_stmt(STMT_RANGE);
	    s1->s.range.to = POP_EXPR();
	    s1->s.range.from = POP_EXPR();
	    if (current_vector == desired_vector &&
		hot_byte >= bv-b.vector &&
		hot_byte <= bv - b.vector +
		b.numbytes_fork + b.numbytes_var_name)
		is_hot_statement = 1;
	    s1->s.range.id.slot =
		READ_BYTES_FROM_VECTOR(bv, b.numbytes_var_name);
	    s1->s.range.id.name = NULL;
	    bv += b.numbytes_var_name;
	    label = READ_BYTES_FROM_VECTOR(bv, b.numbytes_label);
	    PUSH_COND_ARM(alloc_cond_arm(NULL, NULL)); /* dummy */
	    was_hot = is_hot_statement;
	    is_hot_statement = 0;
	    bv = b.vector + stmt2ast(b, bv-b.vector + b.numbytes_label+1,
				  label - b.numbytes_label - 1);
	    if (*bv != OP_JUMP)
		unparse_fail();
	    if (current_vector == desired_vector &&
		hot_byte >= bv - b.vector &&
		hot_byte <= bv - b.vector + b.numbytes_label)
		was_hot = 1;
	    bv += b.numbytes_label;
	    ca1 = POP_COND_ARM();
	    s1->s.range.body = ca1->stmt;
	    myfree(ca1, M_AST);
	    PUSH_STMT(s1, was_hot, TOP);
	    break;
	    
          case OP_FORK:
          case OP_FORK_WITH_ID:
	    old_vector = current_vector;
	    s1 = alloc_stmt(STMT_FORK);
	    s1->s.fork.time = POP_EXPR();
	    label = READ_BYTES_FROM_VECTOR(bv, b.numbytes_fork);
	    if (current_vector == desired_vector &&
		hot_byte >= bv - b.vector &&
		hot_byte <= bv - b.vector + b.numbytes_fork +
		((*bv == OP_FORK_WITH_ID) ? b.numbytes_var_name : 0))
		is_hot_statement = 1;
	    was_hot = is_hot_statement;
	    is_hot_statement = 0;
	    if (*bv == OP_FORK_WITH_ID) {
		bv++;
		s1->s.fork.id.slot =
		    READ_BYTES_FROM_VECTOR(bv, b.numbytes_var_name);
		bv += b.numbytes_var_name;
	    } else {
		s1->s.fork.id.slot = -1;
		bv++;
	    }
	    s1->s.fork.id.name = NULL;
	    PUSH_COND_ARM(alloc_cond_arm(NULL, NULL)); /* dummy */
	    current_vector = label;
	    (void) stmt2ast(prog->fork_vectors[label], 0,
			 prog->fork_vectors[label].size - 1);
	    current_vector = old_vector;
	    ca1 = POP_COND_ARM();
	    s1->s.fork.body = ca1->stmt;
	    myfree(ca1, M_AST);
	    PUSH_STMT(s1, was_hot, TOP);
	    break;
	    
          case OP_JUMP:
	    unparse_fail();
	    return bv-b.vector;
	    break;
	    
          case OP_POP:
	    if (current_vector == desired_vector &&
		hot_byte == bv - b.vector)
		is_hot_statement = 1;
	    s1 = alloc_stmt(STMT_EXPR);
	    s1->s.expr = POP_EXPR();
	    PUSH_STMT(s1, is_hot_statement, TOP);
	    is_hot_statement = 0;
	    break;
	    
          case OP_RETURN:
	    e1 = POP_EXPR(); /* fall through */
	  case OP_RETURN0:   /* e1 defaults to NULL */
	    if (current_vector == desired_vector &&
		hot_byte == bv - b.vector)
		is_hot_statement = 1;
	    s1 = alloc_stmt(STMT_RETURN);
	    s1->s.expr = e1;
	    PUSH_STMT(s1, is_hot_statement, TOP);
	    is_hot_statement = 0;
	    
	  default: ;
	}
	is_hot_statement = 0;
    }
    return bv - b.vector;
}

static int
expr2ast(Bytecodes b, unsigned start, unsigned end)
     /* reads expr's, pushing them onto the expr stack until no */
     /* more expressions can be read or end is reached */
     /* returns index of first unused byte */
{
    Expr *e1 = NULL, *e2, *e3;
    Arg_List *al1;
    unsigned label, label2;
    int i;
    Byte *bv = b.vector + start;
    
    while (bv < b.vector + end) {
	
	if (IS_OPTIM_NUM_OPCODE(*bv)) {
	    i = OPCODE_TO_OPTIM_NUM(*bv);
	    e1 = alloc_var(TYPE_NUM);
	    e1->e.var.v.num = OPCODE_TO_OPTIM_NUM(*bv);
	    PUSH_EXPR(e1);
	}
	
	else if (IS_PUSH_n(*bv)) {
	    e1 = alloc_expr(EXPR_ID);
	    e1->e.id.slot = PUSH_n_INDEX(*bv);
	    e1->e.id.name = NULL;
	    PUSH_EXPR(e1);
	}
	
	else if (IS_PUT_n(*bv)) {
	    e1 = alloc_expr(EXPR_ASGN);
	    e1->e.asgn.kind = ASGN_VAR;
	    e1->e.asgn.a.id.slot = PUT_n_INDEX(*bv);
	    e1->e.asgn.a.id.name = NULL;
	    e1->e.asgn.value = POP_EXPR();
	    PUSH_EXPR(e1);
	}
	
	else if (IS_ARITH_COMP_BIN_OP(*bv)) {
	    e3 = POP_EXPR();
	    e2 = POP_EXPR();
	    switch(*bv) {
	      case OP_ADD:   e1 = alloc_binary(EXPR_PLUS,   e2, e3); break;
	      case OP_MINUS: e1 = alloc_binary(EXPR_MINUS,  e2, e3); break;
	      case OP_MULT:  e1 = alloc_binary(EXPR_TIMES,  e2, e3); break;
	      case OP_DIV:   e1 = alloc_binary(EXPR_DIVIDE, e2, e3); break;
	      case OP_MOD:   e1 = alloc_binary(EXPR_MOD,    e2, e3); break;
	      case OP_EQ:    e1 = alloc_binary(EXPR_EQ,     e2, e3); break;
	      case OP_NE:    e1 = alloc_binary(EXPR_NE,     e2, e3); break;
	      case OP_LT:    e1 = alloc_binary(EXPR_LT,     e2, e3); break;
	      case OP_LE:    e1 = alloc_binary(EXPR_LE,     e2, e3); break;
	      case OP_GT:    e1 = alloc_binary(EXPR_GT,     e2, e3); break;
	      case OP_GE:    e1 = alloc_binary(EXPR_GE,     e2, e3); break;
	      case OP_IN:    e1 = alloc_binary(EXPR_IN,     e2, e3); break;
	    }
	    PUSH_EXPR(e1);
	}
	
	else 
	    switch(*bv) {
	      case OP_AND:
		e1 = alloc_expr(EXPR_AND);
		label = READ_BYTES_FROM_VECTOR(bv, b.numbytes_label);
		e1->e.bin.lhs = POP_EXPR();
		bv = b.vector +
		    expr2ast(b, bv-b.vector + b.numbytes_label + 1, label) - 1;
		e1->e.bin.rhs = POP_EXPR();
		PUSH_EXPR(e1);
		break;
	      case OP_OR:
		e1 = alloc_expr(EXPR_OR);
		label = READ_BYTES_FROM_VECTOR(bv, b.numbytes_label);
		e1->e.bin.lhs = POP_EXPR();
		bv = b.vector +
		    expr2ast(b, bv-b.vector + b.numbytes_label + 1, label) - 1;
		e1->e.bin.rhs = POP_EXPR();
		PUSH_EXPR(e1);
		break;
	      case OP_IMM:
		e1 = alloc_var(TYPE_STR); /* dummy type */
		e1->e.var = vardup(
	            prog->literals[
		        READ_BYTES_FROM_VECTOR(bv, b.numbytes_literal)]);
		bv += b.numbytes_literal;
		PUSH_EXPR(e1);
		break;
	      case OP_G_PUSH:
		e1 = alloc_expr(EXPR_ID);
		e1->e.id.slot =
		    READ_BYTES_FROM_VECTOR(bv, b.numbytes_var_name);
		e1->e.id.name = NULL;
		PUSH_EXPR(e1);
		bv += b.numbytes_var_name;
		break;
	      case OP_G_PUT:   
		e1 = alloc_expr(EXPR_ASGN);
		e1->e.asgn.kind = ASGN_VAR;
		e1->e.asgn.a.id.slot =
		    READ_BYTES_FROM_VECTOR(bv, b.numbytes_var_name);
		e1->e.asgn.a.id.name = NULL;
		e1->e.asgn.value = POP_EXPR();
		PUSH_EXPR(e1);
		bv += b.numbytes_var_name;
		break;
	      case OP_GET_PROP:
	      case OP_PUSH_GET_PROP:
		e3 = POP_EXPR();
		e2 = POP_EXPR();
		e1 = alloc_binary(EXPR_PROP, e2, e3);
		PUSH_EXPR(e1);
		break;
	      case OP_CALL_VERB:
		e3 = POP_EXPR();
		e2 = POP_EXPR();
		e1 = alloc_verb(POP_EXPR(), e2, e3->e.list);
	        myfree(e3, M_AST);
		PUSH_EXPR(e1);
		break;
	      case OP_REF:
	      case OP_PUSH_REF:
		e3 = POP_EXPR();
		e2 = POP_EXPR();
		e1 = alloc_binary(EXPR_INDEX, e2, e3);
		PUSH_EXPR(e1);
		break;
	      case OP_RANGE_REF:
		e1 = alloc_expr(EXPR_RANGE);
		e1->e.range.to = POP_EXPR();
		e1->e.range.from = POP_EXPR();
		e1->e.range.base = POP_EXPR();
		PUSH_EXPR(e1);
		break;
	      case OP_PUT_PROP:
		e1 = alloc_asgn(ASGN_PROP, POP_EXPR());
		e1->e.asgn.a.bin.rhs = POP_EXPR();
		e1->e.asgn.a.bin.lhs = POP_EXPR();
		PUSH_EXPR(e1);
		break;
	      case OP_MAKE_EMPTY_LIST:
		e1 = alloc_expr(EXPR_LIST);
		e1->e.list = NULL;
		PUSH_EXPR(e1);
		break;
	      case OP_MAKE_SINGLETON_LIST:
		e2 = POP_EXPR();
		e1 = alloc_expr(EXPR_LIST);
		e1->e.list = alloc_arg_list(ARG_NORMAL, e2);
		PUSH_EXPR(e1);
		break;
	      case OP_CHECK_LIST_FOR_SPLICE:
		e2 = POP_EXPR();
		e1 = alloc_expr(EXPR_LIST);
		e1->e.list = alloc_arg_list(ARG_SPLICE, e2);
		PUSH_EXPR(e1);
		break;
	      case OP_LIST_ADD_TAIL:
		e3 = POP_EXPR();
		e1 = POP_EXPR();
		al1 = e1->e.list;
		while (al1->next)
		    al1 = al1->next;
		al1->next = alloc_arg_list(ARG_NORMAL, e3);
		PUSH_EXPR(e1);
		break;
	      case OP_LIST_APPEND:
		e3 = POP_EXPR();
		e1 = POP_EXPR();
		al1 = e1->e.list;
		while (al1->next)
		    al1 = al1->next;
		al1->next = alloc_arg_list(ARG_SPLICE, e3);
		PUSH_EXPR(e1);
		break;
	      case OP_BI_FUNC_CALL:
		e3 = POP_EXPR();
		e1 = alloc_expr(EXPR_CALL);
		e1->e.call.args = e3->e.list;
		myfree(e3, M_AST);
		e1->e.call.func = READ_BYTES_FROM_VECTOR(bv, 1);
		PUSH_EXPR(e1);
		bv += 1; 
		break;
	      case OP_NOT:
		e2 = POP_EXPR();
		e1 = alloc_expr(EXPR_NOT);
		e1->e.bin.lhs = e2;
		PUSH_EXPR(e1);
		break;
	      case OP_UNARY_MINUS:
		e1 = POP_EXPR();
		e2 = alloc_binary(EXPR_NEGATE, e1, NULL);
		PUSH_EXPR(e2);
		break;
	      case OP_IF_QUES:
		label = READ_BYTES_FROM_VECTOR(bv, b.numbytes_label);
		if (b.vector[label - b.numbytes_label - 1] != OP_JUMP) {
		    unparse_fail();
		    return bv - b.vector;
		}
		label2 = READ_BYTES_FROM_VECTOR(b.vector +
						label - b.numbytes_label - 1, 
						b.numbytes_label);
		
		e1 = alloc_expr(EXPR_COND);
		e1->e.cond.condition = POP_EXPR();
		bv = b.vector +
		    expr2ast(b, bv - b.vector + b.numbytes_label + 1,
			     label - b.numbytes_label - 1);
		e1->e.cond.consequent = POP_EXPR();
		bv = b.vector + expr2ast(b, label, label2);
		e1->e.cond.alternate = POP_EXPR();
		PUSH_EXPR(e1);
		bv = b.vector + label2 - 1;/* another bv++ at start of loop! */
		break;
	      case OP_PUT_TEMP:
		break;
	      case OP_LISTSET:
		e3 = POP_EXPR();
		e1 = alloc_asgn(ASGN_REF, e3);
		e1->e.asgn.a.bin.rhs = POP_EXPR();
		e1->e.asgn.a.bin.lhs = POP_EXPR();
		PUSH_EXPR(e1);
		for (; *bv != OP_PUSH_TEMP; bv++)
		    ;/*walks subsequent OP_LISTSETs, OP_PUT[PROP] and OP_POP */
		/* now *bv == OP_PUSH_TEMP, which is correct because there
		   will be an additional bv++ in the main loop */
		break;
	      default: /* non-expr constructs -- impossible to proceed */
  		return (bv - b.vector);
	    }
        if (current_vector == desired_vector && hot_statement == NULL &&
	    hot_byte >=start && hot_byte <= (bv - b.vector))
            is_hot_statement = 1;
	bv++;
    }
    return end;
}

void
reset_state()
{
}

/*** external functions ***/
void
list_prog(Program *p, Unparser_Receiver r, void *data, 
	  char user_paren, char user_tab, int f_index) 
     /* char's used as booleans, 1 or 0 */
{
    unsigned i, max_stacks_total = 0;
    Bytecodes bc;
    Stmt *stmt;

    bc = (f_index == MAIN_VECTOR
	  ? p->main_vector
	  : p->fork_vectors[f_index]);
    prog = p;
    receiver = r;
    receiver_data = data;
    hot_byte = 0;

    /* initilizations */
    for (i = 0; i < prog->fork_vectors_size; i++) 
	max_stacks_total += prog->fork_vectors[i].max_stack;
    
    max_stacks_total += prog->main_vector.max_stack;
    
    expr_stack = mymalloc(max_stacks_total * sizeof(Expr *), M_EXPR_STACK);

    reset_state();
    begin_code_allocation();
    stmt = b2ast(bc, 0, bc.size-1);
    /* - 1 corresponding to OP_DONE */
    
    list_prg(stmt, user_paren, user_tab);
    myfree((void *) expr_stack, M_EXPR_STACK);
    free_stmt(stmt);
    
    end_code_allocation(0);
}

static int num_forks;
static int desired_fork;

static int
find_lineno(void *stmt, Stmt *index, unsigned *l, char after_fork)
{
    /*
     * On entry: (*l) is the line number of stmt.
     * On Exit:
     *	  If result == 1, (*l) is the line number of cur_stmt.
     *    Otherwise, (*l) is the line number of whatever follows stmt.
     */
    for (; index; index = index->next) {
	if ((void *) index == stmt) {
	    return 1;
	}
	switch(index->kind) {
	  case STMT_COND:
	    {
		Cond_Arm       *arm;
		int 	       trigger_at_bottom = 0;
		
		for (arm = index->s.cond.arms; arm; arm = arm->next) {
		    if ((void *) arm == stmt)
			if (lp == BOTTOM)
			    trigger_at_bottom = 1;
		        else 
			    return 1;
		    (*l)++;	/* # of consequent of the arm */
		    if (find_lineno(stmt, arm->stmt, l, after_fork))
		        return 1;
		    /* (*l) is # of condition of next arm, if any. */
		}
		if (index->s.cond.otherwise) {
		    (*l)++;    /* for the 'else' */
		    if (find_lineno(stmt, index->s.cond.otherwise, l, after_fork))
			return 1;
		}
		/* (*l) is # of the "endif" line. */
		if (trigger_at_bottom)
		    return 1;
		(*l)++;
	    }
	    break;

	  case STMT_LIST:
	    (*l)++;	/* # of first stmt in body */
	    if (find_lineno(stmt, index->s.list.body, l, after_fork))
		return 1;
	    (*l)++;	/* for the "endfor" line. */
	    break;

	  case STMT_RANGE:
	    (*l)++;	/* # of first stmt in body */
	    if (find_lineno(stmt, index->s.range.body, l, after_fork))
		return 1;
	    (*l)++;	/* for the "endfor" line. */
	    break;

	  case STMT_WHILE:
	    (*l)++;	/* # of first line in body */
	    if (find_lineno(stmt, index->s.loop.body, l, after_fork))
		return 1;
	    (*l)++;	/* for the "endwhile" line. */
	    break;

	  case STMT_FORK:
	    (*l)++;	/* # of first line in body */
	    if (find_lineno(stmt, index->s.fork.body, l, after_fork))
		return 1;
	    if (after_fork != 0 && num_forks == desired_fork)
		return 1;
	    num_forks++;
	    (*l)++;	/* for the "endfork" line */
	    break;

	  case STMT_EXPR:
	  case STMT_RETURN:
	    (*l)++;
	    break;

	  default:
	    errlog("FIND_LINENO: Unknown Stmt_Kind: %d\n", index->kind);
	    break;
	}
    }
    return 0;
}

unsigned
get_lineno(Program *p, unsigned pc, int f_index)
{
    unsigned i, lineno = 1, max_stacks_total = 0;
    Bytecodes bc;
    Stmt *stmt;
    
    bc = (f_index == MAIN_VECTOR
	  ? p->main_vector
 	  : p->fork_vectors[f_index]);
    desired_vector = f_index;
    prog = p;
    hot_byte = pc;

    for (i = 0; i < prog->fork_vectors_size; i++) 
	max_stacks_total += prog->fork_vectors[i].max_stack;
    
    max_stacks_total += prog->main_vector.max_stack;
    
    reset_state();
    num_forks = 0;
    desired_fork = f_index;
    
    if (p->main_vector.size == 1)
	lineno = 0;
    else {
	begin_code_allocation();
	expr_stack = mymalloc(max_stacks_total * sizeof(Expr *), M_EXPR_STACK);
        stmt = b2ast(p->main_vector, 0, p->main_vector.size-1);

        if (hot_statement == NULL && pc != bc.size-1) {
   	    errlog("Couldn't find line number!\n");
 	    lineno = 0;
        } else
	    if (pc == bc.size-1) {
		find_lineno(hot_statement, stmt, &lineno, 1);
		if (desired_fork == MAIN_VECTOR)
		    lineno--;
	    } else 
 	        find_lineno(hot_statement, stmt, &lineno, 0);
	free_stmt(stmt);
	myfree((void *) expr_stack, M_EXPR_STACK);
	end_code_allocation(0);
    }
    return lineno;
}

/* $Log $
   */
